/** @file iop_simplenet.h Simple networked events via tcp/ip */

/* 
   FAU Discrete Event Systems Library (libfaudes)

   Copyright (C) 2008, Thomas Moor
   Exclusive copyright is granted to Klaus Schmidt

*/



#ifndef FAUDES_IOP_SIMPLENET_H
#define FAUDES_IOP_SIMPLENET_H

#include "corefaudes.h"
#include "iop_vdevice.h"


#ifndef FAUDES_WINEXTRA
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <cerrno>
#else
#include <winsock2.h>
#include <pthread.h>
#endif

 

namespace faudes {


/**
 * Simplenet node address.
 *
 * A node address consists of an ip address and a tcp port.
 * This convenience class implements string based access to both components.
 *
 */

class SimplenetAddress {

public:

  /** Default constructor */
  SimplenetAddress(void);

  /** Copy construct */
  SimplenetAddress(const SimplenetAddress& rOther);

  /** Construct from string */
  SimplenetAddress(const std::string& rString);

  /** Return true if valid */
  bool Valid(void) const;

  /** Get ip address */
  std::string Ip(void) const { return mIp; };

  /** Get tcp port */
  int Port(void) const { return mPort; };

  /** Get as colon seperated string */
  std::string IpColonPort(void) const;

  /** Set ip address */
  void Ip(std::string ip) { mIp=ip; };

  /** Set tcp port  */
  void Port(int port) { mPort=port; };

  /** Set from colon seperated string */
  void IpColonPort(std::string ipcolonport);
  
  /** Order for sorting containers of addresses */
  bool operator < (const SimplenetAddress& rOther) const;

 protected:

  /** Ip address */  
  std::string mIp;

  /** Tcp port */
  int mPort;

};


/**
 * Configuration of a network actuator mapping.
 *
 * There is nothing to configure here: any client that
 * connects to this device can subscribe to any actuator event.
 * Future versions of libFAUDES may specify a set of receipent
 * clients.
 *
 */

class AttributeSimplenetActuator : public AttributeVoid {

FAUDES_TYPE_DECLARATION(AttributeSimplenetActuator,AttributeVoid)

public:

  /** Default constructor */
  AttributeSimplenetActuator(void) : AttributeVoid() {}; 

  /** Test for default value (never) */
  virtual bool IsDefault(void) const {return false;};

  /** Clear to default */
  virtual void Clear(void) {};

protected:

  /** Copy method 
   *
   * @param rSrcAttr
   *   Source to copy from
   * @return
   *   Ref to this attribute
   */
  virtual AttributeSimplenetActuator& DoAssign(const AttributeSimplenetActuator& rSrcAttr) 
  {(void) rSrcAttr; return *this;};

  /**
   * Reads the attribute from TokenReader, see AttributeVoid for public wrappers.
   * 
   * If the current token indicates a sensor mapping, the method reads that
   * section. Else it does nothing. Exceptions may only be thrown
   * on invalid data within the section. The label argument is ignored, we use the 
   * hardcoded section "Actuator" for actuator attributes. The context argument is ignored.
   *
   * @param rTr
   *   TokenReader to read from
   * @param rLabel
   *   Section to read
   * @param pContext
   *   Read context to provide contextual information
   *
   * @exception Exception
   *   - IO error (id 1)
   */
  virtual void DoRead(TokenReader& rTr,const std::string& rLabel="", const Type* pContext=0);
 
  /**
   * Writes the attribute to TokenWriter, see AttributeVoid for public wrappers.
   *
   * Writes the actuator mapping data. The label argument is ignored, we use 
   * the hardcoded section "Actuator". The context argument is ignored.  
   *
   * @param rTw
   *   TokenWriter to write to
   * @param rLabel
   *   Section to write
   * @param pContext
   *   Read context to provide contextual information
   *
   * @exception Exception
   *   - IO error (id 2)
   */
  virtual void DoWrite(TokenWriter& rTw,const std::string& rLabel="", const Type* pContext=0) const;

}; // end class AttributeSimplenetActuator


/**
 * Configuration of a network sensor mapping.
 *
 * There is nothing to configure here: any server that
 * provides the repective event will be accepted. Future versions 
 * of libFAUDES may specify a set of servers here.
 *
 */

class AttributeSimplenetSensor : public AttributeVoid {

FAUDES_TYPE_DECLARATION(AttributeSimplenetSensor,AttributeVoid)

public:

  /** Default constructor (no triggers) */
 AttributeSimplenetSensor(void) : AttributeVoid() {}; 

  /** Test for default value (never) */
  virtual bool IsDefault(void) const {return false;};

  /** Clear to default */
  virtual void Clear(void) {};

protected:

  /** Copy method 
   *
   * @param rSrcAttr
   *   Source to copy from
   * @return
   *   Ref to this attribute
   */
  virtual AttributeSimplenetSensor& DoAssign(const AttributeSimplenetSensor& rSrcAttr) 
  {(void) rSrcAttr; return *this;};


  /**
   * Reads the attribute from TokenReader, see AttributeVoid for public wrappers.
   * 
   * If the current token indicates a sensor mapping, the method reads that
   * section. Else it does nothing. Exceptions may only be thrown
   * on invalid data within the section. The label argument is ignored, we the hardcoded
   * section "Sensor" for sensor device attributes. The context argument is ignored.
   *
   * @param rTr
   *   TokenReader to read from
   * @param rLabel
   *   Section to read
   * @param pContext
   *   Read context to provide contextual information
   *
   * @exception Exception
   *   - IO error (id 1)
   */
  virtual void DoRead(TokenReader& rTr,const std::string& rLabel="", const Type* pContext=0);
 
  /**
   * Writes the attribute to TokenWriter, see AttributeVoid for public wrappers.
   *
   * Writes the sensor mapping data.The label argument is ignored, we use 
   * the hardcoded section "Sensor". The context argument is ignored.  
   *
   * @param rTw
   *   TokenWriter to write to
   * @param rLabel
   *   Section to write
   * @param pContext
   *   Read context to provide contextual information
   *
   * @exception Exception
   *   - IO error (id 2)
   */
  virtual void DoWrite(TokenWriter& rTw,const std::string& rLabel="", const Type* pContext=0) const;

}; // end class AttributeSimplenetSensor



/**
 * Configuration of a networked sensor or actuator 
 *
 * This class is derived from the AttributeDeviceEvent to specialise
 * for networked sensor and actuator mapping.
 *
 */

class AttributeSimplenetEvent : public AttributeDeviceEvent {

 FAUDES_TYPE_DECLARATION(AttributeSimplenetEvent,AttributeDeviceEvent)

public:

  /** Default constructor (no mapping at all) */
  AttributeSimplenetEvent(void);

  /** Copy constructor */
  AttributeSimplenetEvent(const AttributeSimplenetEvent& rOtherAttr);

  /** Test for default value (never) */
  virtual bool IsDefault(void) const {return false;};

  /** Clear */
  virtual void Clear(void)  { AttributeDeviceEvent::Clear(); };

  /**  Get actuator mapping */
  const AttributeSimplenetActuator*  Actuatorp(void) const {
    return static_cast<AttributeSimplenetActuator*>(mpActuatorAttribute); };

  /**  Get sensor mapping */
  const AttributeSimplenetSensor*  Sensorp(void) const {
    return static_cast<AttributeSimplenetSensor*>(mpSensorAttribute); };

protected:

  /** DoAssign */
  virtual AttributeSimplenetEvent& DoAssign(const AttributeSimplenetEvent& rSrc)  
  { AttributeDeviceEvent::DoAssign(rSrc); return *this;};

  /** Prototype, sensor */
  static AttributeSimplenetSensor sSensorPrototype;

  /** Prototype, actuator */
  static AttributeSimplenetActuator sActuatorPrototype;

}; // class AttributeSimplenetEvent


/**
 * An nDevice implements networked IO via a simple tcp/ip protocol. 
 *
 * @section SecIodeviceNDev1 Networking
 *
 * The purpose of this implementation of networked discrete events is to provide
 * a basis for case studies. However, simple networked events via nDevice must not 
 * be confused with ongoing research that addresses synchronisation and real 
 * time behaviour. We take a really pragmatic approach here.
 *
 * Our <em>network</em> is identified by its <em>network name</em>. It consists
 * a number of <em>nodes</em>, each identified by its <em>node name</em>. In the current implementation,
 * each node knows about the entire network ie knows about all other nodes. Each node plays two roles:
 * - the <em>server</em> role to send event notifications, and
 * - the <em>client</em> role to receive event notifications.
 *
 * As a server, the node is configured to listen on its <em>server port</em> for incomming 
 * tcp connections. When a client (eg some other node in client role) connects to the tcp port, the server 
 * replies to a simple command set in order to allow the client to subscribe to the nodes actuator events.
 * When the application executes an actuator on the node, this is notified to all connected clients and
 * the clients will interpret the event as a sensor reading. Vice versa, the node itself has the 
 * role of a client and subscribes its sensor events from all other nodes. 
 *
 * @section SecIodeviceNDev2 Protocol Details
 *
 * The protocol for commands and notification is libFAUDES-token based and can be inspected eg via nc or telnet; 
 * see the shell script in the tutorial. 
 * Event notifications are sent spontaneous to all subscribing nodes in the format 
 * <tt>\<Notify\> "event" \</Notify\></tt>.
 * Commands are accepted in the format <tt>\<Cmd\> "command" \</Cmd\></tt> and will be
 * answered accordingly.
 *
 * <table>
 * <tr>
 * <td> Command on tcp server port</td>
 * <td> Reply </td>
 * </tr>
 * <tr>
 * <td>  <tt>\<Cmd\> "Info"  \</Cmd\></tt> </td>
 * <td> Configuration of node as token string </td>
 * </tr>
 * <tr>
 * <td> <tt>\<Cmd\> "Status"  \</Cmd\></tt>  </td>
 * <td> Up/StartUop/Down/ShutDown </td>
 * </tr>
 * <tr>
 * <td> <tt>\<Subscribe\> "ev1" ... "evN"  \</Subscribe\></tt> </td>
 * <td> Positive Acknowledge to accept subscription of events </td>
 * </tr>
 * </table>
 *
 * @section SecIodeviceNDev3 Name Resolution
 *
 * On the technical side, each node needs to figure the ip addresses incl tcp ports on which the 
 * other nodes provide their service. To ease configuration, this information is distributed by 
 * udp datagrams. In order to locate the other nodes within the network, a node may  broadcasts a 
 * network <em>request</em> datagramm. Each node that receives such a request, replies with an 
 * <em>advert</em> datagramm to provide its address. Thus, the simple net nodes rely on some 
 * underlying name resolution by dns services, but node configuration itself refers to simple-net node names
 * only. Since each node knows about the names of all participating nodes, each node will know when 
 * all connections are up. 
 * 
 * <table>
 * <tr>
 * <td> Broadcast on simplenet udp port</td>
 * <td> Reply </td>
 * </tr>
 * <tr>
 * <td> <tt>\<Request\> "network"  \</Request\></tt> </td>
 * <td> Advertisement of networkname, nodename and ip-address:port </td>
 * </tr>
 * </table>
 * <p>
 *
 * @section SecIodeviceNDev4 File IO
 *
 * For token IO, the nDevice reads and writes a section with label  
 * "SimplenetDevice". There are no relevant attributes yet. Simple machine example:
 *
 * @code
 * <SimplenetDevice>
 *
 * "simplemachine"     % node id
 * 1000                % time scale (ms/ftu)
 * "localhost:40000"   % server tcp port 
 * "simpleloop"        % network id
 *
 * <Network>           
 * "simplemachine"     % other nodes ..,
 * "simplesupervisor"  % ... in this network
 * </Network>
 *
 * <SimplenetEvents>
 * "alpha"  <Input></Input>
 * "beta"   <Output></Output>
 * "mue"    <Output></Output>
 * "lambda" <Input></Input>
 * </SimplenetEvents>
 *
 * </SimplenetDevice>
 * @endcode
 *
 * @section SecIodeviceNDev5 Implementation Notes
 *
 * The current status of the code is premature; network io
 * assumes reasonably large buffers; thread/select mechanism is inefficient; 
 * exception handling wont work; etc etc
 * 
 *
 * @ingroup IODevicePlugin 
 */

class nDevice : public vDevice {    

  // provide all access to background tasks
  friend void* NDeviceListen(void*);
  friend void* NDeviceServer(void*);
  friend void* NDeviceReply(void*);

 public:
  /**
   * Default constructor
   */
  nDevice(void);

  /**
   * Explicit destructor.
   */
  virtual ~nDevice(void);

  /**
   * Clear all configuration.
   * This implies Stop().
   */
  virtual void Clear(void);

  /**
   * Set up internal data structures
   *
   */
  virtual void Compile(void);

  /**
   * Add entry to node name resolution
   *
   * @param rNode
   *   Name of node to register
   * @param rAddress 
   *  Address incl port 
   */
  void InsertNodeAddress(const std::string& rNode, const SimplenetAddress& rAddress);
 

  /**
   * Activate the device. This function enables actuator execution and sensor reading.
   * It starts the background thread for incomming connections and tries to connect to relevant servers.
   *
   * @exception Exception
   *   - Not yet configured (id 551)
   *   - Fatal network error (id 553)
   *   - Fatal pthread error (id 554)
   */
  virtual void Start(void);

  /**
   * Deactivate the device. This function disables actuator execution and sensor reading.
   * It stops the backhround thread to provide connections and disconnects from any servers.
   */
  virtual void Stop(void);

  /**
   * Reset device. Resets buffered sensor events and current time.
   *
   */
  virtual void Reset(void);

  /**
   * Run actuator command.
   *
   * @exception Exception
   *   - unknown actuator event (id 65)
   */
  virtual void WriteActuator(Idx actuator);



 protected:

  /**
   * Actual method to read device configuration from tokenreader.
   *
   * This method calls the base class to read the device name and the
   * the timescale. It then reads address, networkid and nodes.
   * Overall device configuration is consists of DoReadPreface,
   * DoReadConfiguration and Compile. It isimplemented in vDevice.
   * The label and context parameters are ignored.
   *
   * @param rTr
   *   TokenReader to read from
   * @param rLabel
   *   Section to read
   * @param pContext
   *   Read context to provide contextual information
   *
   * @exception Exception
   *   - IO error (id 1)
   */
  virtual void DoReadPreface(TokenReader& rTr,  const std::string& rLabel = "", const Type* pContext=0);
 
  /**
   * Actual method to write the device configuration to a TokenWriter.
   *
   * This method calls the base class vDevice to write the device name and
   * the time scale. It then writes network related data. The label and
   * context parameters are ignored.
   *
   * @param rTw
   *   Reference to TokenWriter
   * @param rLabel
   *   Label of section to write
   * @param pContext
   *   Read context to provide contextual information
   * @exception Exception 
   *   - IO errors (id 2)
   */
  virtual void DoWritePreface(TokenWriter& rTw, const std::string& rLabel="", const Type* pContext=0) const;


  /** Overall configuration (with actual type) */
  TaNameSet<AttributeSimplenetEvent>* pConfiguration ;

  /** Simplenet: network id */
  std::string mNetwork;

  /** Simplenet: address of my server port */
  SimplenetAddress mAddress;

  /** Simplenet: effective address of my server port */
  SimplenetAddress mEffectiveAddress;

  /** Simplenet: list of nodes in this network */
  std::set<std::string> mNetworkNodes;
  
  /** Simplenet: list of event server addresses */
  std::set<std::string> mSensorServers;
  
  /** Compiled data: map subscriptions */
  std::map<std::string,EventSet> mSensorSubscriptions;

  /** Background: mutex for below shared variables*/
  pthread_mutex_t mMutex;

  /** Background: thread handle (global) */
  pthread_t mThreadListen; 

  /** Background: request to join via flag (mutexed) */
  bool mStopListen;

  /** Background: server socket to listen (background only) */
  int mServerSocket;
  
  /** Background: server port to bind (background only) */
  struct sockaddr_in mServerAddress;

  /** Background: udp broadcast socket (background only) */
  int mBroadcastSocket;
  
  /** Background: state of a connection to a client (shared) */
  typedef struct {
    nDevice* mDevice;        // refernce to this device
    int  mClientSocket;      // the socket the client is connected to
    pthread_t mThread;       // the thread we run to reply to client requests
    EventSet mEvents;        // events the client has subscribed to 
  } ClientState;  

  /** Background: map sockets to connection states (shared) */
  std::map<int, ClientState> mActuatorClientStates;

  /** Background: state of a connection to an event server (shared)*/
  typedef struct {
    SimplenetAddress mAddress;  // ip address incland tcp port of remote server
    EventSet mEvents;           // events we have subscribed to
    int  mServerSocket;         // socket used to connect with provider
    std::string mLineBuffer;    // buffer to receive one line
  } ServerState;  

  /** Background: connection states to event servers (by node name)*/
  std::map<std::string,ServerState> mSensorServerStates;



}; // end class nDevice

// declare background threads
void* NDeviceListen(void*);
void* NDeviceServer(void*);
void* NDeviceReply(void*);

} // namespace faudes


#endif // include
