/** @file iop_simplenet.cpp Simple networked events via tcp/ip */

/* 
   FAU Discrete Event Systems Library (libfaudes)

   Copyright (C) 2008, Thomas Moor
   Exclusive copyright is granted to Klaus Schmidt

*/



#include "iop_simplenet.h"
#include <time.h>
#include <sys/time.h>


// fix winsock 
#ifdef FAUDES_WINEXTRA
#define close closesocket
#define socklen_t int
#define WINCASTSTR (char *)
#else
#define WINCASTSTR 
#endif

// fix non strict bsd
#ifndef SO_REUSEPORT
#define SO_REUSEPORT SO_REUSEADDR
#endif

// todo: runtime opt for broadcast port
#define FAUDES_IODEVICE_BROADCASTPORT 40000

namespace faudes {


/*
 **********************************************
 **********************************************
 **********************************************

 implementation: SimplenetAddress

 **********************************************
 **********************************************
 **********************************************
 */

// construct
SimplenetAddress::SimplenetAddress(void) {
  mIp="";
  mPort=-1;
}

// copy construct
SimplenetAddress::SimplenetAddress(const SimplenetAddress& rOther) {
  mIp=rOther.mIp;
  mPort=rOther.mPort;
}

// construct from string
SimplenetAddress::SimplenetAddress(const std::string& rString) {
  IpColonPort(rString);
}

// validity
bool SimplenetAddress::Valid(void) const {
  if(mPort<=0) return false;
  if(mIp=="") return false;
  if(mIp.find(':',0)!=std::string::npos) return false;
  return true;
}  

// get colon seperated string 
std::string SimplenetAddress::IpColonPort(void) const {
  std::string res;
  if(!Valid()) return res;
  res=mIp+":"+ToStringInteger(mPort);
  return res;
}

// Set from colon seperated string 
void SimplenetAddress::IpColonPort(std::string ipcolonport) {
  FD_DHV("SimplenetAddress::IpColonPort(): " << ipcolonport << " --> ?");
  // invalid
  mIp="";
  mPort=-1;
  // find colon
  unsigned int cpos = ipcolonport.find(':',0);
  if(cpos==std::string::npos) return;
  if(cpos==0) return;
  if(cpos+1>= ipcolonport.length()) return;
  // extract ip
  mIp=ipcolonport.substr(0,cpos);
  // extract port
  mPort=ToIdx(ipcolonport.substr(cpos+1));
  // report
  FD_DHV("SimplenetAddress::IpColonPort(): " << ipcolonport << " --> " << IpColonPort());
  // test for errors (might be too strict)
  if(IpColonPort()!=ipcolonport) {
    mIp="";
    mPort=-1;
  }
}
  
// sorting
bool SimplenetAddress::operator < (const SimplenetAddress& rOther) const {
  if(this->mIp < rOther.mIp) return true;
  if(this->mIp > rOther.mIp) return false;
  if(this->mPort < rOther.mPort) return true;
  return false;
}

/*
 **********************************************
 **********************************************
 **********************************************

 implementation: AttributeSimplenetActuator

 **********************************************
 **********************************************
 **********************************************
 */

// std faudes type
FAUDES_TYPE_IMPLEMENTATION(AttributeSimplenetActuator,AttributeVoid,)


//DoWrite(rTw);
void AttributeSimplenetActuator::DoWrite(TokenWriter& rTw, const std::string& rLabel, const Type* pContext) const {
  (void) rLabel; (void) pContext;
  rTw.WriteBegin("Output");
  rTw.WriteEnd("Output");
}


//DoRead(rTr)
void AttributeSimplenetActuator::DoRead(TokenReader& rTr, const std::string& rLabel, const Type* pContext) {
  (void) rLabel; (void) pContext;
  // report
  FD_DHV("AttributeSimplenetActuator(" << this << ")::DoRead(tr)");

  // test for actuator section
  Token token;
  rTr.Peek(token);
  if(token.Type()!=Token::Begin) return;
  if(token.StringValue()!="Output") return;

  // read actuator section
  rTr.ReadBegin("Output");
  rTr.ReadEnd("Output");
}


/*
 **********************************************
 **********************************************
 **********************************************

 implementation: AttributeSimplenetSensor

 **********************************************
 **********************************************
 **********************************************
 */


// std faudes type
FAUDES_TYPE_IMPLEMENTATION(AttributeSimplenetSensor,AttributeVoid,)


//DoWrite(rTw);
void AttributeSimplenetSensor::DoWrite(TokenWriter& rTw, const std::string& rLabel, const Type* pContext) const {
  (void) rLabel; (void) pContext;
  rTw.WriteBegin("Input");
  rTw.WriteEnd("Input");
}


//DoRead(rTr)
void AttributeSimplenetSensor::DoRead(TokenReader& rTr, const std::string& rLabel, const Type* pContext) {
  (void) rLabel; (void) pContext;
  // report
  FD_DHV("AttributeSimplenetSensor(" << this << ")::DoRead(tr)");

  // test for sensor section
  Token token;
  rTr.Peek(token);
  if(token.Type()!=Token::Begin) return;
  if(token.StringValue()!="Input") return;

  // read sensor section
  rTr.ReadBegin("Input");
  rTr.ReadEnd("Input");
}




/*
 **********************************************
 **********************************************
 **********************************************

 implementation: AttributeSimplenetEvent

 **********************************************
 **********************************************
 **********************************************
 */


// statics
AttributeSimplenetActuator AttributeSimplenetEvent::sActuatorPrototype;
AttributeSimplenetSensor  AttributeSimplenetEvent::sSensorPrototype;

// std faudes type
FAUDES_TYPE_IMPLEMENTATION(AttributeSimplenetEvent,AttributeDeviceEvent,)

// Default constructor, set my prototypes
AttributeSimplenetEvent::AttributeSimplenetEvent(void) : AttributeDeviceEvent() {
  FD_DHV("AttributeSimplenetEvent::AttributeSimplenetEvent(" << this << ")");;
  pActuatorPrototype=&sActuatorPrototype;
  pSensorPrototype=&sSensorPrototype;
}

// Copy constructor 
AttributeSimplenetEvent::AttributeSimplenetEvent(const AttributeSimplenetEvent& rOtherAttr) : 
  AttributeDeviceEvent() 
{
  FD_DHV("AttributeSimplenetEvent(" << this << "): form other attr " <<  &rOtherAttr);
  pActuatorPrototype=&sActuatorPrototype;
  pSensorPrototype=&sSensorPrototype;
  DoAssign(rOtherAttr);
}



/*
 **********************************************
 **********************************************
 **********************************************

 implementation: nDevice

 **********************************************
 **********************************************
 **********************************************
 */

// helper: send entire buffer
int syncSend(int dest, const char* data, int len, int flag) {
  int from=0;
  int left=len;
  while(left>0) {
     int rc=send(dest, data+from, left, 0);
     if(rc<0) {
       std::stringstream errstr;
       errstr << "Simplenet fatal network error (cannot send message)";
       throw Exception("nDevice::syncSend", errstr.str(), 553);
     }
    left-=rc;
  }
  return len;
}


// constructor
nDevice::nDevice(void) : vDevice() {
  FD_DH("nDevice(" << this << ")::nDevice()");
  mpConfiguration = new TaNameSet<AttributeSimplenetEvent>;
  pConfiguration = dynamic_cast< TaNameSet<AttributeSimplenetEvent>* >(mpConfiguration);
  // default token section 
  mDefaultLabel="SimplenetDevice";
  // default network data
  mName="SimplenetNode";
  mNetwork="Simplenet";
  mAddress.IpColonPort("localhost:40000");
  // clear/prepare state
  pthread_mutex_init(&mMutex, NULL); 	
  mServerSocket=-1;
  mBroadcastSocket=-1;
}

// destructor
nDevice::~nDevice(void) {
  FD_DH("nDevice(" << this << ")::~nDevice()");
  // stop
  Stop();
  // clean up my data
  pthread_mutex_destroy(&mMutex);
}

// clear all configuration
void nDevice::Clear(void) {
  FD_DH("nDevice(" << this << ")::Clear()");
  // call base, incl stop
  vDevice::Clear();
  // clear compiled data
  mSensorSubscriptions.clear();
  mSensorServers.clear();
}

// add node address
void nDevice::InsertNodeAddress(const std::string& node, const SimplenetAddress& rAddress) {  
  FD_DHV("nDevice(" << this << ")::InsertNodeAddress(" << node << ", " << rAddress.IpColonPort() << ")");
  std::map<std::string,nDevice::ServerState>::iterator sit;
  sit=mSensorServerStates.find(node);
  // bail out on unknown nodes
  if(sit==mSensorServerStates.end()) return;
  // bail out on knwon address
  if(sit->second.mAddress.Valid()) return;
  // record
  sit->second.mAddress=rAddress;
  // todo: fix active sockets
  sit->second.mServerSocket=-1;
  // done
  FD_DH("nDevice(" << this << ")::InsertNodeAddress(" << node << ", " << rAddress.IpColonPort() << "): accept as new");
}


//Compile(void)
void nDevice::Compile(void){
  //setup up internal data structure
  FD_DHV("nDevice(" << this << ")::Compile()");
  // call base
  vDevice::Compile();
}


//DoWritePreface(rTr,rLabel)
void nDevice::DoWritePreface(TokenWriter& rTw, const std::string& rLabel, const Type* pContext) const {
  FD_DHV("vDevice::DoWrite()");
  //call base
  vDevice::DoWritePreface(rTw,rLabel,pContext);
  // write my data
  rTw.WriteString(mAddress.IpColonPort());
  rTw.WriteString(mNetwork);
  // write my nodes
  rTw.WriteBegin("Network");
  std::set<std::string>::const_iterator nit;
  for(nit=mNetworkNodes.begin();nit!=mNetworkNodes.end();nit++) 
    rTw.WriteString(*nit);
  rTw.WriteEnd("Network");
}

//DoRead(rTr,rLabel)
void nDevice::DoReadPreface(TokenReader& rTr, const std::string& rLabel, const Type* pContext) {
  FD_DHV("vDevice::DoReadPreface()");
  // call base (reads name and timescale)
  vDevice::DoReadPreface(rTr,rLabel,pContext);
  // read my data: server port
  mAddress.IpColonPort(rTr.ReadString());
  if(!mAddress.Valid()) {
    std::stringstream errstr;
    errstr << "Simplenet address expected at " << rTr.FileLine();
    throw Exception("nDevice::DoRead", errstr.str(), 50);
  }
  // read my data: network id
  mNetwork=rTr.ReadString();
  // read network nodes
  mNetworkNodes.clear();
  rTr.ReadBegin("Network");
  while(!rTr.Eos("Network")) {
    mNetworkNodes.insert(rTr.ReadString());
  }
  rTr.ReadEnd("Network");
}


// lock - unlock shortcuts;
#define LOCK_E  {int rc = pthread_mutex_lock(&mMutex); \
  if(rc) {FD_ERR("nDevice::LOCK_E: lock mutex error\n"); exit(1); }}		
#define UNLOCK_E  {int rc = pthread_mutex_unlock(&mMutex); \
  if(rc) {FD_ERR("nDevice::LOCK_E: unlock mutex error\n"); exit(1); }}		
#define TLOCK_E  {int rc = pthread_mutex_lock(&ndevice->mMutex); \
  if(rc) {FD_ERR("nDevice::TLOCK_E: lock mutex error\n"); exit(1); }}		
#define TUNLOCK_E  {int rc = pthread_mutex_unlock(&ndevice->mMutex); \
  if(rc) {FD_ERR("nDevice::TLOCK_E: unlock mutex error\n"); exit(1); }}		


// Write Actuator 
void nDevice::WriteActuator(Idx actuator) {

  FD_DH("vDevice::WriteActuator(" << mActuators.SymbolicName(actuator) << ")");

  // bail out (do notify clients even when servers incomplete)
  if(mState!=Up && mState!=StartUp) return;

  // test event
  if(!mActuators.Exists(actuator)) {
    std::stringstream errstr;
    errstr << "Unknown actuator event " << actuator;
    throw Exception("nDevice::WriteActuator", errstr.str(), 65);
  }

  // find properties
  const AttributeSimplenetActuator* aattr = pConfiguration->Attribute(actuator).Actuatorp();
  if(!aattr) {
    std::stringstream errstr;
    errstr << "Invalid actuator attribute " << actuator;
    throw Exception("nDevice::WriteActuator", errstr.str(), 65);
  }

  // report
  std::string message= "<Notify> \"" + mActuators.SymbolicName(actuator) + "\" </Notify>\n";
  FD_DH("vDevice::WriteActuator(): message: " << message.substr(0,message.length()-1));
  
  // send event to those clients that did subscribe 
  LOCK_E;
  std::map<int,ClientState>::iterator sit=mActuatorClientStates.begin();
  for(;sit!=mActuatorClientStates.end();sit++) {
    if(!sit->second.mEvents.Empty())
      if(!sit->second.mEvents.Exists(actuator))
        continue;
    int clientsock=sit->second.mClientSocket;
    FD_DH("vDevice::WriteActuator(): to socket " << clientsock);
    syncSend(clientsock, message.c_str(), message.length(), 0);
  }
  UNLOCK_E;
  FD_DH("vDevice::WriteActuator(): done");
}



// Start(void)
void nDevice::Start(void) {
  if(mState!=Down) return;
  FD_DH("nDevice(" << mName <<")::Start()");
  // call base
  vDevice::Start();
  mState=StartUp;
  // clear event server states
  mSensorServerStates.clear();
  std::set<std::string>::iterator nit;
  for(nit=mNetworkNodes.begin(); nit!=mNetworkNodes.end();nit++) {
    if(*nit == mName) continue;
    mSensorServerStates[*nit].mAddress= SimplenetAddress("unknown:0");
    mSensorServerStates[*nit].mEvents= EventSet();
    mSensorServerStates[*nit].mServerSocket=-1;
    mSensorServerStates[*nit].mLineBuffer="";
  }
  // set my effective address
  char hostname[1024];
  int hostname_len =1023;
  if(gethostname(hostname,hostname_len)!=0) {
    std::stringstream errstr;
    errstr << "Simplenet fatal network error (cannot get hostname)";
    throw Exception("nDevice::Start", errstr.str(), 553);
  }
  hostname[hostname_len]=0;
  mEffectiveAddress=mAddress;
  mEffectiveAddress.Ip(hostname);
  FD_DH("nDevice::Start(): server adress " << mEffectiveAddress.IpColonPort());
  // open a tcp port to listen: create socket
  mServerSocket = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
  if(mServerSocket<=0) {
    std::stringstream errstr;
    errstr << "Simplenet fatal network error (cannot open server socket)";
    throw Exception("nDevice::Start", errstr.str(), 553);
  }
  int reuse=1;
  setsockopt(mServerSocket,SOL_SOCKET, SO_REUSEADDR, WINCASTSTR &reuse, sizeof(reuse));  
  // open a tcp port to listen: set up address
  memset(&mServerAddress, 0, sizeof(mServerAddress));
  mServerAddress.sin_family = AF_INET;
  mServerAddress.sin_addr.s_addr = htonl(INADDR_ANY);  
  mServerAddress.sin_port = htons(mAddress.Port());     
  // open a tcp port to listen: bind socket to address
  if(bind(mServerSocket, (struct sockaddr *) &mServerAddress,sizeof(mServerAddress)) <0) {
    std::stringstream errstr;
    errstr << "Simplenet fatal network error (cannot bind socket)";
    throw Exception("nDevice::Start", errstr.str(), 553);
  }
  // open a tcp port to listen: start to listen
  if(listen(mServerSocket, 77) < 0) {  // todo: max pending connections
    std::stringstream errstr;
    errstr << "Simplenet fatal network error (cannot listen from socket)";
    throw Exception("nDevice::Start", errstr.str(), 553);
  }
  // open a udp port to listen: create socket
  mBroadcastSocket = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
  if(mBroadcastSocket<=0) {
    std::stringstream errstr;
    errstr << "Simplenet fatal network error (cannot open broadcast socket)";
    throw Exception("nDevice::Start", errstr.str(), 553);
  }
  //int reuse=1;
  setsockopt(mBroadcastSocket,SOL_SOCKET, SO_REUSEADDR, WINCASTSTR &reuse, sizeof(reuse));  
  setsockopt(mBroadcastSocket,SOL_SOCKET, SO_REUSEPORT, WINCASTSTR &reuse, sizeof(reuse));  
  // open a udp port: enable broadcast 
  // int reuse
  if(setsockopt(mBroadcastSocket, SOL_SOCKET, SO_BROADCAST, (char *) &reuse, sizeof(reuse)) ) {
    std::stringstream errstr;
    errstr << "Simplenet fatal network error (cannot setopt broadcast socket)";
    throw Exception("nDevice::Start", errstr.str(), 553);
  }  
  // open a udp port to listen: set up address
  struct sockaddr_in broadcastaddr;
  memset(&broadcastaddr, 0, sizeof(broadcastaddr));
  broadcastaddr.sin_family = AF_INET;
  broadcastaddr.sin_addr.s_addr = htonl(INADDR_ANY);  
  broadcastaddr.sin_port = htons(FAUDES_IODEVICE_BROADCASTPORT);     
  // open a udp port to listen: bind socket to address
  if(bind(mBroadcastSocket, (struct sockaddr *) &broadcastaddr,sizeof(broadcastaddr)) <0) {
    std::stringstream errstr;
    errstr << "Simplenet fatal network error (cannot bind broadcast socket)";
    throw Exception("nDevice::Start", errstr.str(), 553);
  }
  // start background thread to listen: create thread
  pthread_attr_t attr;	
  pthread_attr_init(&attr);
  pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
  // start background threadto listen: run thread
  mStopListen=false;
  int rc  = pthread_create(&mThreadListen, &attr, NDeviceListen, this);
  // start background thread to listen: dont need the attribute anymore
  pthread_attr_destroy(&attr);
  // pthread error 
  if(rc) {
    std::stringstream errstr;
    errstr << "Simplenet fatal pthread error (cannot create thread)";
    throw Exception("nDevice::Start", errstr.str(), 554);
  }
}

// Stop(void)
void nDevice::Stop(void) {
  // bail out
  if(mState!=Up && mState!=StartUp) return;
  FD_DH("nDevice::Stop()");
  // stop background threads
  LOCK_E;
  mStopListen=true;
  UNLOCK_E;
  // signal update to my threads: via udp message ... release select
  std::string message= "<Stop> \"" + mNetwork + "\" \"" + mName + "\" </Stop>\n";
  struct sockaddr_in broadcastaddr;
  broadcastaddr.sin_family=AF_INET;
  broadcastaddr.sin_port=htons(FAUDES_IODEVICE_BROADCASTPORT);
  broadcastaddr.sin_addr.s_addr=htonl(INADDR_BROADCAST);
  LOCK_E;
  sendto(mBroadcastSocket,message.c_str(),message.length(),0,(struct sockaddr *) & broadcastaddr, sizeof(broadcastaddr));
  UNLOCK_E;
  // wait until listen thread finished ... implies all other threads finished too
  FD_DH("nDevice::Stop(): waiting for listen thread");
  pthread_join(mThreadListen, NULL);
  // close server socket
  shutdown(mServerSocket,2);
  close(mServerSocket);
  mServerSocket=-1;
  // close broadcast socket
  shutdown(mBroadcastSocket,2);
  close(mBroadcastSocket);
  mBroadcastSocket=-1;
  // call base
  vDevice::Stop();
}



// background thread, listening on my port for clients to connect
void* NDeviceListen(void* arg){
  bool term;
  std::map<std::string,nDevice::ServerState>::iterator sit;
  std::map<int,nDevice::ClientState>::iterator cit;
  // cast this object
  nDevice* ndevice= static_cast<nDevice*>(arg);
  // say hello
  FD_DH("nDevice::Listen(" << ndevice << ")");
  // clear broadcast time stamp
  timeval lastbroadcast; 	
  lastbroadcast.tv_sec=0;
  lastbroadcast.tv_usec=0;

  // infinite loop 
  while(true){

    // report missing servers
    int missing=0;
    for(sit=ndevice->mSensorServerStates.begin(); sit!=ndevice->mSensorServerStates.end(); sit++) {
      if(sit->second.mServerSocket>0) continue;
      FD_DH("nDevice::Listen(): missing server for node " << sit->first);
      missing++;
    }
    if(missing>0 && ndevice->mState==vDevice::Up) {
      TLOCK_E;
      ndevice->mState=vDevice::StartUp;
      TUNLOCK_E;
    }
    if(missing==0 && ndevice->mState==vDevice::StartUp) {
      TLOCK_E;
      ndevice->mState=vDevice::Up;
      TUNLOCK_E;
    }
    if(missing>0 && ndevice->mState==vDevice::StartUp) {
      // is a broadcast due?
      timeval now; 	
      gettimeofday(&now,0);
      if(now.tv_sec - lastbroadcast.tv_sec >5) {
        // udp message
        std::string message= "<Request> \"" + ndevice->mNetwork + "\" </Request>\n";
        // udp broadcast
        struct sockaddr_in broadcastaddr;
        broadcastaddr.sin_family=AF_INET;
        broadcastaddr.sin_port=htons(FAUDES_IODEVICE_BROADCASTPORT);
        broadcastaddr.sin_addr.s_addr=htonl(INADDR_BROADCAST);
        TLOCK_E;
        int rc=sendto(ndevice->mBroadcastSocket,message.c_str(),message.length(),0,(struct sockaddr *) & broadcastaddr, sizeof(broadcastaddr));
        (void) rc;
        FD_DH("nDevice::Listen(): broadcast message: " <<  message.substr(0,message.length()-1) << " #" << rc);
        TUNLOCK_E;
        gettimeofday(&lastbroadcast,0);
      }
    }

    // initiate/maintaine subscriptions
    for(sit=ndevice->mSensorServerStates.begin(); sit!=ndevice->mSensorServerStates.end(); sit++) {
      // have active connection?
      if(sit->second.mServerSocket>0) continue;
      // have an address?
      if(!sit->second.mAddress.Valid()) continue;
      // try to connect
      FD_DH("nDevice::Listen(): subscribing to " << sit->first << 
          " at " << sit->second.mAddress.IpColonPort());
      // open a tcp port: create socket
      int serversock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
      if(serversock<=0) {
        FD_DH("nDevice::Listen(): subscription failed: no socket");
        continue;
      }
      // open a tcp port: set up internet address
      unsigned long int serverinaddr = INADDR_NONE;
      if(serverinaddr==INADDR_NONE) {
        FD_DH("nDevice::Listen(): using address in dot format");
        serverinaddr = inet_addr(sit->second.mAddress.Ip().c_str());
      }
      if(serverinaddr==INADDR_NONE) {
        struct hostent *host;
        host = gethostbyname(sit->second.mAddress.Ip().c_str());
        if(host!=0) {
          FD_DH("nDevice::Listen(): unsing address by name lookup");
          serverinaddr = *(unsigned long int*) host->h_addr;
        }
      }
      if(serverinaddr==INADDR_NONE) {
        FD_DH("nDevice::Listen(): subscription failed: invalid address " << sit->second.mAddress.Ip());
        close(serversock);
        continue;
      }
      // open a tcp port: set up socket address 
      struct sockaddr_in serveraddress;
      memset(&serveraddress, 0, sizeof(serveraddress));
      serveraddress.sin_family = AF_INET;
      serveraddress.sin_addr.s_addr=serverinaddr; 
      serveraddress.sin_port = htons(sit->second.mAddress.Port());     
      // open a tcp port: connect
      if(connect(serversock, (struct sockaddr*) &serveraddress, sizeof(serveraddress))<0) {
        FD_DH("nDevice::Listen(): subscription failed: connect");
        close(serversock);
        continue;
      }    
      // record success, say hello to remote sensor server, ask for info
      FD_DH("nDevice::Listen(): subscribing to " << sit->first << " via socket " << serversock);
      sit->second.mServerSocket=serversock;
      std::string hello;
      hello="% Simplenet universal event subscription: \""+ndevice->mName+"\" subscribing from \""+sit->first+"\"\n"; 
      syncSend(serversock, hello.c_str(), hello.length(), 0);
      hello="% Expecting notifications in format '<Notify> \"event_name\" </Notify>'\n";
      syncSend(serversock, hello.c_str(), hello.length(), 0);
      hello="% Sending info command, explicit subscription may follow\n";
      syncSend(serversock, hello.c_str(), hello.length(), 0);
      hello="<Cmd> \"Info\" </Cmd>\n";
      syncSend(serversock, hello.c_str(), hello.length(), 0);
      FD_DH("nDevice::Listen(): subscribing to " << sit->first << " via socket " << serversock << ": ok");
    }

    // prepare relevant  on sources ... 
    fd_set mysocks;
    int mysocks_max=0;
    FD_ZERO(&mysocks);
    // ... my server listen socket , expecting other nodes to subscribe
    FD_SET(ndevice->mServerSocket, &mysocks);
    if(mysocks_max<ndevice->mServerSocket) mysocks_max=ndevice->mServerSocket;
    // ... udp port, expecting requests and adverts 
    FD_SET(ndevice->mBroadcastSocket, &mysocks);
    if(mysocks_max< ndevice->mBroadcastSocket) mysocks_max=ndevice->mBroadcastSocket;
    // ... sensor server connections, expecting notifications
    for(sit=ndevice->mSensorServerStates.begin(); sit!=ndevice->mSensorServerStates.end(); sit++) {
      int serversock=sit->second.mServerSocket;
      if(serversock<0) continue;
      FD_SET(serversock, &mysocks);
      if(mysocks_max< serversock) mysocks_max=serversock;
    }

    // wait for traffic with moderate timeout 
    struct timeval tv;
    tv.tv_sec = 10;
    tv.tv_usec = 0;
    FD_DH("nDevice::Listen(): listen as node \"" << ndevice->mName << "\" on network \"" << ndevice->mNetwork << "\"");
    int avail=select(mysocks_max+1, &mysocks, NULL, NULL, &tv);

    // handle request for subscription
    if(avail>0)  if(FD_ISSET(ndevice->mServerSocket,&mysocks)) {
      int clientsock=-1;
      struct sockaddr_in clientaddr;
      socklen_t clientaddr_len = sizeof(clientaddr);
      clientsock=accept(ndevice->mServerSocket, (struct sockaddr *) &clientaddr, &clientaddr_len );
      if(clientsock<0) {
        FD_DH("nDevice::Listen(): failed to accept incomming connection");
        break;
      }
      FD_DH("nDevice::Listen(): connection request from " << inet_ntoa(clientaddr.sin_addr));
      // lock for client data base
      TLOCK_E;
      // record client
      nDevice::ClientState* cstate = &ndevice->mActuatorClientStates[clientsock];
      cstate->mDevice=ndevice;
      cstate->mClientSocket=clientsock;
      // start server thread for new client
      pthread_attr_t attr;	
      pthread_attr_init(&attr);
      pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
      pthread_t clientthread;
      int rc = pthread_create(&clientthread, &attr, NDeviceReply,cstate);
      pthread_attr_destroy(&attr);
      cstate->mThread=clientthread;
      // unlock for client data base
      TUNLOCK_E;
      // pthread error 
      if(rc) {
        std::stringstream errstr;
        errstr << "Simplenet fatal pthread error (cannot create thread)";
        throw Exception("nDevice::Listen()", errstr.str(), 554);
      }
    }

    // handle incomming broadcast
    if(avail>0)  
    if(FD_ISSET(ndevice->mBroadcastSocket,&mysocks)) {
      // get message
      char data[1024]; 
      int data_len=1023;
      struct sockaddr_in fromaddr;
      socklen_t fromaddr_len = sizeof(fromaddr);
      data_len=recvfrom(ndevice->mBroadcastSocket,data,data_len,0, (struct sockaddr*) &fromaddr,&fromaddr_len);
      if(data_len<0) data_len=0; // todo:  eof
      data[data_len]=0;
      if(data_len>=1) if(data[data_len-1]=='\n') data[data_len-1]=0;
      FD_DH("nDevice::Listen(): received udp datagram " << data << 
         " from " << inet_ntoa(fromaddr.sin_addr));
      // interpret message
      TokenReader tr(TokenReader::String,std::string(data));
      try {
        Token token;
        tr.Peek(token);
        // interpret udp request ...
        if(token.Type()==Token::Begin && token.StringValue()=="Request") {
	  tr.ReadBegin("Request");
          if(tr.ReadString()==ndevice->mNetwork) {
            // set up advert
            std::string message= "<Advert> \"" 
                + ndevice->mNetwork + "\" \"" 
                + ndevice->mName    + "\"" +
 	        " \"" + ndevice->mEffectiveAddress.IpColonPort() +"\"" +  
                "  </Advert>\n"; 
            FD_DH("nDevice::Listen(): reply advert: " <<  message.substr(0,message.length()-1));
            // udp broadcast
            struct sockaddr_in broadcastaddr;
            broadcastaddr.sin_family=AF_INET;
            broadcastaddr.sin_port=htons(FAUDES_IODEVICE_BROADCASTPORT);
            broadcastaddr.sin_addr.s_addr=htonl(INADDR_BROADCAST);
            TLOCK_E;
            sendto(ndevice->mBroadcastSocket,message.c_str(),message.length(),0,(struct sockaddr *) & broadcastaddr, sizeof(broadcastaddr));
	    TUNLOCK_E
          } else {
            FD_DH("nDevice::Listen(): ingoring request from other network");
	  }
	}
        // interpret udp advert ...
        if(token.Type()==Token::Begin && token.StringValue()=="Advert") {
	  tr.ReadBegin("Advert");
          if(tr.ReadString()==ndevice->mNetwork) {
            std::string node = tr.ReadString();
            std::string saddr = tr.ReadString();
            SimplenetAddress addr(saddr);
            addr.Ip(inet_ntoa(fromaddr.sin_addr)); // overwrite with actual ip
            FD_DH("nDevice::Listen(): accepting advert from my network");
            ndevice->InsertNodeAddress(node,addr);
	  } else {
            FD_DH("nDevice::Listen(): ingoring advert from other network");
          }
        }
      } catch (faudes::Exception&) {
        FD_DH("nDevice::Listen(): invalid udp message");
      }
    }
       

    // handle event notification and other commands from remote sensor servers
    if(avail>0)  
    if(!FD_ISSET(ndevice->mServerSocket,&mysocks)) 
    if(!FD_ISSET(ndevice->mBroadcastSocket,&mysocks)) {
      for(sit=ndevice->mSensorServerStates.begin(); sit!=ndevice->mSensorServerStates.end(); sit++) {
        int serversock=sit->second.mServerSocket;
	if(serversock<0) continue;
	if(FD_ISSET(serversock, &mysocks)) {
          FD_DH("nDevice::Listen(): reading sock " <<  serversock);
          // buffer data in line buffer
          char buffer[1025];
          int count = recv(serversock, buffer, 1024, 0);
          if(count<=0) { // todo: test eof
            FD_DH("nDevice::Listen(): reading sock " <<  serversock << " : eof");
            close(serversock); 
            sit->second.mServerSocket=-1;
            continue;
          } 
          FD_DH("nDevice::Listen(): reading sock " <<  serversock  << ": #" << count);
          buffer[count]=0;
          sit->second.mLineBuffer +=std::string(buffer);
          // interpret line(s)
          if(count>0) 
          if(buffer[count-1]=='\n') 
          if(sit->second.mLineBuffer.length()>0) 
          {
            const std::string& linebuffer = sit->second.mLineBuffer;
            FD_DH("nDevice::Listen(): " <<  serversock  << ": line: " << linebuffer);
            // tokenise notification
            TokenReader tr(TokenReader::String,linebuffer);
            int ncount=0;
            try {
              Token token;
              while(tr.Peek(token)) {
                // its an event notify
                if(token.Type()==Token::Begin && token.StringValue()=="Notify") {
                  tr.ReadBegin("Notify");
  	          std::string event = tr.ReadString();
                  tr.ReadEnd("Notify");
                  pthread_mutex_lock(ndevice->pBufferMutex);
                  FD_DH("nDevice::Listen(): found event " << event);
                  Idx sev=ndevice->mSensors.Index(event);
                  if(sev) ndevice->pSensorBuffer->push_back(sev);
                  pthread_mutex_unlock(ndevice->pBufferMutex);
                  ncount++;
                }
                // its an info
                if(token.Type()==Token::Begin && token.StringValue()=="SimplenetDevice") {
                  FD_DH("nDevice::Listen(): found device info");
		  nDevice remote;
                  remote.Read(tr);
                  FD_DH("nDevice::Listen(): found device with actuators " << remote.Actuators().ToString());      
                  EventSet sevents=ndevice->Sensors();
                  sevents.SetIntersection(remote.Actuators());
		  sevents.Name("Subscribe");
		  std::string message=sevents.ToString();
                  FD_DH("nDevice::Listen(): subscribing events " << message);
                  message += "\n";
                  syncSend(serversock,message.c_str(), message.length(),0);
                }
              }
            } catch (faudes::Exception&) {
              FD_DH("nDevice::Listen(): " <<  serversock  << ": invalid notification");
	    }
            sit->second.mLineBuffer.clear();
            // signal condition
            if(ncount>0) {
              FD_DH("nDevice::Listen(): broadcast condition");
              pthread_mutex_lock(ndevice->pWaitMutex);
              pthread_cond_broadcast(ndevice->pWaitCondition);
              pthread_mutex_unlock(ndevice->pWaitMutex);
            }
	  }
	}
      }
    }

    // some error
    if(avail<0) {
      FD_DH("nDevice::Listen(): select error");
    }

    // terminate thread on request
    TLOCK_E;
    term= ndevice->mStopListen;
    TUNLOCK_E; 
    if(term) break;
  }

  // enforce reply threads to terminate
  FD_DH("nDevice::Listen(): cancel reply threads");
  TLOCK_E;
  for(cit=ndevice->mActuatorClientStates.begin(); cit!=ndevice->mActuatorClientStates.end(); cit++) {
    pthread_cancel(cit->second.mThread);
  }
  TUNLOCK_E; 
  FD_DH("nDevice::Listen(): close reply sockts");
  TLOCK_E;
  for(cit=ndevice->mActuatorClientStates.begin(); cit!=ndevice->mActuatorClientStates.end(); cit++) {
    int clientsock= cit->second.mClientSocket;
    if(clientsock>0) close(clientsock);
    cit->second.mClientSocket=-1;
  }
  TUNLOCK_E; 
  FD_DH("nDevice::Listen(): discard client data");
  TLOCK_E;
  ndevice->mActuatorClientStates.clear();
  TUNLOCK_E; 
  // done
  FD_DH("nDevice::Listen(): terminating listen thread");
  for(sit=ndevice->mSensorServerStates.begin(); sit!=ndevice->mSensorServerStates.end(); sit++) {
    int serversock=sit->second.mServerSocket;
    if(serversock>0) close(serversock);
    sit->second.mServerSocket=-1;
  }
  pthread_exit(NULL);
  return NULL;
}





// background thread, reply on client request
void* NDeviceReply(void* arg) {
  bool term=false;
  // cast my state
  nDevice::ClientState* cstate= static_cast<nDevice::ClientState*>(arg);
  nDevice* ndevice = cstate->mDevice;
  // get my socket
  TLOCK_E;
  int clientsock = cstate->mClientSocket;
  TUNLOCK_E; 
  FD_DH("nDevice::Reply(" <<  clientsock  << "): start");
  // say hello
  std::string hello;
  hello="% Simplenet Event Server: \""+ndevice->mName+"\" providing events\n";
  send(clientsock, hello.c_str(), hello.length(), 0);
  hello="% Notifications will have format '<Notify> \"event_name\" </Notify>'\n";
  syncSend(clientsock, hello.c_str(), hello.length(), 0);
  hello="% Commands are accepted in format '<Cmd> \"cmd_name\" </Cmd>'\n";
  syncSend(clientsock, hello.c_str(), hello.length(), 0);
  hello="% Supported commands are \"Info\" and \"Status\"\n";
  syncSend(clientsock, hello.c_str(), hello.length(), 0);
  // have a linebuffer
  std::string linebuffer;
  // loop forever
  while(true) {
    FD_DH("nDevice::Reply(" <<  clientsock  << "): waiting");
    // allow for cancel
    pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);
    // select my socket with moderate time out
    fd_set mysock;
    FD_ZERO(&mysock);
    FD_SET(clientsock, &mysock);
    struct timeval tv;
    tv.tv_sec = 100;
    tv.tv_usec = 0;
    pthread_testcancel(); // fix LinuxThread were select is no cancelation point
    int avail=select(clientsock+1, &mysock, NULL, NULL, &tv);
    pthread_testcancel();
    // dont cancle
    pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL);
    // read data to buffer
    FD_DH("nDevice::Reply(" <<  clientsock  << "): release");
    if(avail>0) {
      char buffer[1025];
      int count = recv(clientsock, buffer, 1024, 0);
      if(count<=0) { // todo: test eof
        FD_DH("nDevice::Reply(" <<  clientsock  << "): eof");
        term=true;
        break;
      } 
      FD_DH("nDevice::Reply(" <<  clientsock  << "): count: " << count);
      buffer[count]=0;
      linebuffer +=std::string(buffer);
      avail=0;
      // pass on complete lines for interpretation
      if(count>0) if(buffer[count-1]=='\n') avail=1;
    }
    // evaluate linebuffer
    if(avail>0) {
      FD_DH("nDevice::Reply(" <<  clientsock  << "): line: " << linebuffer);
      // tokenise command
      TokenReader tr(TokenReader::String,linebuffer);
      try {
        Token token;
        while(tr.Peek(token)) {
          // its a command
          if(token.Type()==Token::Begin && token.StringValue()=="Cmd") { 
  	    tr.ReadBegin("Cmd");
	    std::string cmd = tr.ReadString();
  	    tr.ReadEnd("Cmd");
            std::string response="<NAck> </NAck>\n";
            // command: info
            if(cmd=="Info") {
              TLOCK_E;
              response=ndevice->ToString() + "\n";
              TUNLOCK_E;
            }
            // command: status
            if(cmd=="Status") {
              TLOCK_E;
              if(ndevice->mState==vDevice::Up) response="<Ack> \"Up\" </Ack>\n";
              if(ndevice->mState==vDevice::StartUp) response="<Ack> \"StartUp\" </Ack>\n";
              if(ndevice->mState==vDevice::ShutDown) response="<Ack> \"ShutDown\" </Ack>\n";
              TUNLOCK_E;
            }
            // send reply
            syncSend(clientsock, response.c_str(), response.length(), 0);
	  }
          // its a event subscription
          if(token.Type()==Token::Begin && token.StringValue()=="Subscribe") { 
	    EventSet sevents;
            sevents.Read(tr,"Subscribe");
            FD_DH("nDevice::Reply(" <<  clientsock  << "): providing events " << sevents.ToString());
            TLOCK_E;
            cstate->mEvents.InsertSet(sevents);
            TUNLOCK_E;
	  }
	}
      } catch (faudes::Exception&) {
        FD_DH("nDevice::Reply(" <<  clientsock  << "): invalid cmd");
      }
      // done
      linebuffer="";
    }
    // lost connection / socket closed
    if(avail<0) {
      FD_DH("nDevice::Reply(" <<  clientsock  << "): socket closed");
      break;
    }
  } // loop
  // remove client state
  TLOCK_E;
  if(cstate->mClientSocket>0) close(clientsock);
  cstate->mClientSocket=-1;
  ndevice->mActuatorClientStates.erase(clientsock); 
  TUNLOCK_E;
  // done
  FD_DH("nDevice::Reply(" << clientsock << "): terminate");
  pthread_exit(NULL);
  return NULL;
}

// reset dynamic faudes state (buffered events, current time)
void nDevice::Reset(void) {
  // call base for time and sensor buffer
  vDevice::Reset();
}



} // name space



