/** @file graphfncts.h Operations on (directed) graphs. */

/* FAU Discrete Event Systems Library (libfaudes)

   Copyright (C) 2009  Thomas Moor, Klaus Schmidt, Sebastian Perk
   Exclusive copyright is granted to Klaus Schmidt

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#ifndef FAUDES_GRAPHFNCTS_H

#include "definitions.h"
#include "vgenerator.h"
#include <stack>

namespace faudes {


/**
 * Filter for strictly connected components (SCC) search/compute routines.
 *
 * When ComputeScc()/HasScc() traverse a specified transition sytem in their
 * search, a SccFilter parameter can mute certain transitions. The respective
 * filter conditions are set by the constructor and consist of a flag word (mMode)
 * and additional parameters (depending on the flags). The Flags can be combined
 * from the following: 
 *
 * - NoFilter: std textbook beheviour;
 * - FindFirst: stop the search after one SCC has been found;
 * - IgnoreTrivial: ignore SCCs that consist of one state without selfloop;
 * - StatesAvoid: ignore SCCs that contain one or more states from specified set;
 * - StatesRequire: ignore SCCs that fail to contain at least one state froma specified set;
 * - EventsAvoid: mute any transitions with an event label from the specified set.
 *
 * Convenience modes set up the state related filter conditions from the set of marked states:
 *
 * - IgnoreLiveLocks: set StatesRequire to the marked states of a specified Generator;
 * - LiveLocksOnly: set StatesAvoid to the marked states of a specified Generator;
 * - IgnoreUnaccessible: initialise todo list with accessible states only.
 *
 * Technical note: SccFilters are static objects in that they can not be edited after 
 * construction; interally, they use references to condition parameters that are required to
 * exisit during the life time of the filter object. 
 *
 * Technical note: currently, there is no EventsRequire filter because the implementation 
 * interprets the transition relation primarily from a directed graph perspective; while StatesRequire
 * makes sense for marked states semantics, we are not aware of any applications for
 * a corresponding version of EventsRequire; please let us know if you require such an extension. 
 *
 * @ingroup GeneratorFunctions
 */
class SccFilter {

  // Allow search function access
  friend void SearchScc(
    const Idx vState,           
    int& vRcount,               
    const Generator& rGen,      
    const SccFilter& rFilter,
    StateSet&  rTodo,           
    std::stack<Idx>& rStack,    
    StateSet& rStackStates,     
    std::map<const Idx, int>& rDfn,
    std::map<const Idx, int>& rLowLnk, 
    std::list<StateSet>& rSccList,  
    StateSet& rRoots);       

public:

  /** Typedef for filter modes */
  typedef enum { 
    NoFilter=0x00,             //// no filters at all (default)
    FindFirst=0x01,            //// only report first valid SCC
    IgnoreTrivial=0x02,        //// ignore non-cycles, aka singletons without selfloop
    StatesAvoid=0x10,          //// filter by avoiding specified states
    StatesRequire=0x20,        //// filter by requireing specified states
    EventsAvoid=0x40,          //// filter by avoiding specified events
    IgnoreLiveLocks=0x1000,    //// ignore cycles with unmarked states only
    LiveLocksOnly =0x2000,     //// ignore cycles one or more marked states
    IgnoreUnaccessible=0x4000  //// only inspect accessible part
  } FMode;

  /** Constructor (no filter) */
  SccFilter(void);

  /** Constructor (from flags w.r.t. Generator) */
  SccFilter(int mode, const Generator& rGen);

  /** Constructor (from flags and state sets) */
  SccFilter(int mode, const StateSet& rStatesAvoid, const StateSet& rStatesRequire);

  /** Constructor (from flags and event sets) */
  SccFilter(int mode, const EventSet& rEventsAvoid);

  /** Constructor (from flags and sets) */
  SccFilter(int mode, const StateSet& rStatesAvoid, const StateSet& rStatesRequire, 
     const EventSet& rEventsAvoid);

  /** Destructor */
  ~SccFilter(void);

  /** Member access */
  int Mode(void) const { return mMode;};

protected:

  /** Flag, combining bit masks from Mode */
  int mMode; 
  /** States to avoid (if flag StatesAvoid is set) */
  const StateSet* pStatesAvoid;
  /** States to require (if flag StatesRequire is set) */
  const StateSet* pStatesRequire;
  /** Events to avoid (if flag EventssAvoid is set) */
  const EventSet* pEventsAvoid;

  /** Local sets (optional) */
  StateSet* mpStatesAvoid;
  StateSet* mpStatesRequire;
  EventSet* mpEventsAvoid;

  /** Static emptysets */
  const static StateSet msEmptyStates;
  const static EventSet msEmptyEvents;
};


/**
 *
 * Search for strongly connected components (SCC). 
 * 
 * This function partitions the stateset of a generator into equivalent 
 * classes such that states x1 and x2 are equivalent iff there is a path from x1
 * to x2 AND a path from x2 to x1. 
 *
 * This function implements the algorithm based on a depth first search
 * presented in:
 *
 * -- Aho, Hopcroft, Ullman: The Design and Analysis of Computer Algorithms --
 *
 * While the original algorithm works on a directed graph, this
 * implementation adds some features that refer to transition systems and
 * allow to filter SCCs on the run. The filter condition is specified by the
 * SccFilter parameter rFilter.
 *
 * Note: this version is derived from earlier implementations used in 
 * various plug-ins; in due course, this version will replace earlier versions; 
 * filter conditions are not thouroughly tested.
 *
 * @param vState
 *   State, from which the current recursion is started.
 * @param vRcount
 *   Denotes the current depth of the recursion.
 * @param rGen
 *   Transition system to investigate
 * @param rFilter
 *   Filter out specified transitions
 * @param rTodo
 *   Set of states that up to now were not found by the
 *   depth first search.
 * @param rStack
 *   Stack of states to represent current path.
 * @param rStackStates
 *   Set of states that are in rStack
 * @param rDfn
 *   Map assigning to each state idx its Depth-First Number.
 * @param rLowLnk
 *   Map assigning to each state its LOWLINK Number.
 * @param rSccList
 *   Set SCCs (accumulative result).
 * @param rRoots
 *   Set of states that each are root of some SCC (accumulative result).
 *
 * 
 */
void SearchScc(
  const Idx vState,           
  int& vRcount,               
  const Generator& rGen,      
  const SccFilter& rFilter,
  StateSet&  rTodo,           
  std::stack<Idx>& rStack,    
  StateSet& rStackStates,     
  std::map<const Idx, int>& rDfn,
  std::map<const Idx, int>& rLowLnk, 
  std::list<StateSet>& rSccList,  
  StateSet& rRoots);       


/**
 * Compute strongly connected components (SCC) 
 *
 * This partitions the stateset 
 * of a generator into equivalenc classes such that states x1 and x2 are equivalent 
 * iff there is a path from x1 to x2 AND a path from x2 to x1. You may specify 
 * the filter condition rFilter to ignore certain transitions and/or SCCs; see also SccFilter. 
 *
 * Technically, this function is a API wrapper that calls the recursive implementation
 * SearchScc() as presented in 
 *
 * -- Aho, Hopcroft, Ullman: The Design and Analysis of Computer Algorithms --
 *
 * 
 * @param rGen
 *   Generator under investigation
 * @param rFilter
 *   Filter out specified transitions
 * @param rSccList
 *   List of SCCs (result)
 * @param rRoots
 *   Set of states that each are root of some SCC (result).
 *
 * @return
 *   True if SCCs have been found, false if not.
 *
 * @ingroup GeneratorFunctions
 * 
 */
bool ComputeScc(
  const Generator& rGen,
  const SccFilter& rFilter,
  std::list<StateSet>& rSccList,
  StateSet& rRoots);

/**
 * Compute strongly connected components (SCC) 
 *
 * This partitions the stateset 
 * of a generator into equivalenc classes such that states x1 and x2 are equivalent 
 * iff there is a path from x1 to x2 AND a path from x2 to x1. 
 *
 * Technically, this function is a API wrapper that calls the recursive implementation
 * SearchScc() as presented in
 * 
 * -- Aho, Hopcroft, Ullman: The Design and Analysis of Computer Algorithms --
 *
 *
 * @param rGen
 *   Generator under investigation
 * @param rSccList
 *   List of SCCs (result)
 * @param rRoots
 *   Set of states that each are root of some SCC (result).
 *
 * @return
 *   True if SCCs have been found, false if not.
 *   Since there are no filters, true is returned iff the
 *   the state set is non-empty.
 *
 * @ingroup GeneratorFunctions
 * 
 */
bool ComputeScc(
  const Generator& rGen,
  std::list<StateSet>& rSccList,
  StateSet& rRoots);
  

  
/**
 * Test for strongly connected components (SCC) 
 *
 * This functions searchs for the first SCC of the generator rGen 
 * while applying the filter rFilter; see SCCFilter for details.
 *
 * Technically, this function is a API wrapper that calls the recursive implementation
 * SearchScc() as presented in 
 *
 * -- Aho, Hopcroft, Ullman: The Design and Analysis of Computer Algorithms --
 * 
 * @param rGen
 *   Generator under investigation
 * @param rFilter
 *   Filter out specified transitions
 * @param rScc
 *   First SCC that has been found, empty if no such.
 *
 * @return
 *   True if SCCs have been found, false if not.
 *
 * @ingroup GeneratorFunctions
 * 
 */
bool HasScc(
  const Generator& rGen,
  const SccFilter& rFilter,
  StateSet& rScc
);

/**
 * Test for strongly connected components (SCC) 
 *
 * This functions searchs for the first SCC of the generator rGen 
 * while applying the filter rFilter; see SCCFilter for details.
 *
 * Technically, this function is a API wrapper that calls the recursive implementation
 * SearchScc() as presented in 
 * 
 * -- Aho, Hopcroft, Ullman: The Design and Analysis of Computer Algorithms --
 *
 * @param rGen
 *   Generator under investigation
 * @param rFilter
 *   Filter out specified transitions
 *
 * @return
 *   True if SCCs have been found, false if not.
 *
 * @ingroup GeneratorFunctions
 * 
 */
bool HasScc(
  const Generator& rGen,
  const SccFilter& rFilter
);

} // namespace
#endif
