/** @file parallel.h parallel composition */

/* FAU Discrete Event Systems Library (libfaudes)

   Copyright (C) 2006  Bernd Opitz
   Exclusive copyright is granted to Klaus Schmidt

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#ifndef FAUDES_PARALLEL_H

#include "agenerator.h"
#include <stack>
#include <map>
#include <set>

namespace faudes {


/**
 * Parallel composition.
 *
 * Constructs the parallel composition of two generators, where shared events
 * are synchronised while non-shared events are executed independantly. 
 * The resulting generators alphabet is the union of the argument alphabets.
 * In this implementation, only accessible states are generated. 
 * On deterministic input this functions constructs a deterministic output.
 * See also Parallel(const vGenerator&,std::map< std::pair<Idx,Idx>, Idx>&,const vGenerator&, vGenerator&).
 *
 * @param rGen1
 *   First generator
 * @param rGen2
 *   Second generator
 * @param rResGen
 *   Reference to resulting parallel composition generator
 *
 * <h4>Example:</h4>
 * <table border=0> <tr> <td> <table>
 * <tr> <td> Generator G1 </td> <td> Generator G2 </td> </tr>
 * <tr>
 * <td> @image html tmp_parallel_g1.png </td>
 * <td> @image html tmp_parallel_g2.png </td>
 * </tr>
 * </table> </td> </tr> <tr> <td> <table width=100%>
 * <tr> <td> G1 || G2 </td> </tr>
 * <tr> <td> @image html tmp_parallel_g1g2.png </td> </tr>
 * </table> </td> </tr> </table>
 *
 * @ingroup GeneratorFunctions
 *
 */
void Parallel(const vGenerator& rGen1, const vGenerator& rGen2, vGenerator& rResGen);


/**
 * Parallel composition.
 *
 * See also Parallel(const vGenerator&, const vGenerator&, vGenerator&).
 * This version tries to be transparent on event attributes: if
 * argument attributes match and if the result can take the respective
 * attributes, then they are copied; it is considered an error if 
 * argument attributes do not match.
 *
 * @param rGen1
 *   First generator
 * @param rGen2
 *   Second generator
 * @param rResGen
 *   Reference to resulting composition generator
 *
 * @ingroup GeneratorFunctions
 */
void aParallel(const vGenerator& rGen1, const vGenerator& rGen2, vGenerator& rResGen);

/**
 * Parallel composition.
 *
 * See Parallel(const vGenerator&, const vGenerator&, vGenerator&).
 * This version fills given a reverse composition map to map pairs of old states
 * to new states.
 *
 * @param rGen1
 *   First generator
 * @param rGen2
 *   Second generator
 * @param rReverseCompositionMap
 *   Reverse composition map (map< pair<Idx,Idx>, Idx>)
 * @param rResGen
 *   Reference to resulting parallel composition generator
 */
void Parallel(
    const vGenerator& rGen1, const vGenerator& rGen2,
    std::map< std::pair<Idx,Idx>, Idx>& rReverseCompositionMap, 
    vGenerator& rResGen);


/**
 * Product composition.
 * 
 * The product composition executes shared events only. The resulting
 * generators alphabet is the interscetion of the argument alphabets.
 * In this implementation, only accessible states are generated. 
 * Assumes deterministic input generators, result is deterministic.
 *
 * @param rGen1
 *   First generator
 * @param rGen2
 *   Second generator
 * @param rResGen
 *   Reference to resulting product composition generator
 *
 * @ingroup GeneratorFunctions
 */
void Product(const vGenerator& rGen1, const vGenerator& rGen2, vGenerator& rResGen);


/**
 * Product composition.
 * 
 * See Product(const vGenerator&, const vGenerator&, vGenerator&).
 * This version fills given reverse composition map to map pairs of old states
 * to new states.
 *
 * @param rGen1
 *   First generator
 * @param rGen2
 *   Second generator
 * @param rReverseCompositionMap
 *   Reverse composition map (map< pair<Idx,Idx>, Idx>)
 * @param rResGen
 *   Reference to resulting product composition generator
 */
void Product(
    const vGenerator& rGen1, const vGenerator& rGen2,
    std::map< std::pair<Idx,Idx>, Idx>& rReverseCompositionMap, 
    vGenerator& rResGen);


/**
 * Product composition.
 *
 * See also Product(const vGenerator&, const vGenerator&, vGenerator&).
 * This version tries to be transparent on event attributes: if
 * argument attributes match and if the result can take the respective
 * attributes, then they are copied; it is considered an error if 
 * argument attributes do not match.
 *
 * @param rGen1
 *   First generator
 * @param rGen2
 *   Second generator
 * @param rResGen
 *   Reference to resulting product composition generator
 *
 * @ingroup GeneratorFunctions
 */
void aProduct(const vGenerator& rGen1, const vGenerator& rGen2, vGenerator& rResGen);


/**
 * Helper: uses reverse composition map to track state names
 * in a paralell composition. Purely cosmetic.
 *
 * @param rGen1
 *   First generator
 * @param rGen2
 *   Second generator
 * @param rReverseCompositionMap
 *   Reverse composition map (map< pair<Idx,Idx>, Idx>)
 * @param rGen12
 *   Reference to resulting parallel composition generator
 */
void SetComposedStateNames(
    const vGenerator& rGen1, const vGenerator& rGen2,
    const std::map< std::pair<Idx,Idx>, Idx>& rReverseCompositionMap, 
    vGenerator& rGen12);

/**
 * Helper: extract composition map from reverse composition map.
 * The reverse composition map is an optional return value of functions
 * that operate on the product state space of two component generators, e.g. Parallel. 
 * It maps pairs of component state indices to the correponding index in the product set. 
 * Occasionally, one requires a map from the product space to to one of the component indicees.
 * This function extracts the latter from the reverse composition map. There are two versions,
 * one for each component.
 * 
 * @param rReverseCompositionMap
 *  provided map from pairs of component states to states in the product space
 * @param rCompositionMap
 *  resulting map from the product space to states from the first compoent
 * 
 */
void CompositionMap1(
  const std::map< std::pair<Idx,Idx>, Idx>& rReverseCompositionMap, 
  std::map<Idx,Idx>& rCompositionMap);


/**
 * Helper: extract composition map from reverse composition map.
 * See also CompositionMap1.
 * 
 * @param rReverseCompositionMap
 *  provided map from pairs of component states to states in the product space
 * @param rCompositionMap
 *  resulting map from the product space to states from the second component
 * 
 */
void CompositionMap2(
  const std::map< std::pair<Idx,Idx>, Idx>& rReverseCompositionMap, 
  std::map<Idx,Idx>& rCompositionMap);



/**
 * Parallel composition with relaxed acceptance condition.
 *
 * This version of the parallel composition relaxes the synchronisation of the acceptance
 * condition (marking). It requires that the omega extension of the generated language
 * has infinitely many prefixes that comply to the marked languages of G1 and G2, each.
 * It does however not require the synchronous acceptance.
 *
 * @param rGen1
 *   First generator
 * @param rGen2
 *   Second generator
 * @param rResGen
 *   Reference to resulting parallel composition generator
 *
 *
 * @ingroup GeneratorFunctions
 *
 */
void OmegaParallel(const vGenerator& rGen1, const vGenerator& rGen2, vGenerator& rResGen);


/**
 * Parallel composition with relaxed acceptance condition.
 *
 * See also OmegaParallel(const vGenerator&, const vGenerator&, vGenerator&).
 * This version tries to be transparent on event attributes: if
 * argument attributes match and if the result can take the respective
 * attributes, then they are copied; it is considered an error if 
 * argument attributes do not match.
 *
 * @param rGen1
 *   First generator
 * @param rGen2
 *   Second generator
 * @param rResGen
 *   Reference to resulting composition generator
 *
 * @ingroup GeneratorFunctions
 */
void aOmegaParallel(const vGenerator& rGen1, const vGenerator& rGen2, vGenerator& rResGen);

} // namespace faudes

#define FAUDES_PARALLEL_H
#endif 

