/** @file regular.cpp

Operations on regular languages.
See [Cassandras and Lafortune. Introduction to Discrete Event Systems] for an
introduction to regular language operations.
Operations are always performed on language(s) marked by the passed generator(s),
resulting in the language(s) marked by the resulting generator(s).
Only if mentioned extra, the same is done for the involved generated (prefix-closed)
languages.

*/

/* FAU Discrete Event Systems Library (libfaudes)

Copyright (C) 2006  Bernd Opitz
Exclusive copyright is granted to Klaus Schmidt

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */

  
#include "regular.h"


namespace faudes {

// LanguageUnionNonDet(rGen1, rGen2, rResGen)
void LanguageUnionNonDet(const vGenerator& rGen1, const vGenerator& rGen2,
			 vGenerator& rResGen) {
            
  FD_DF("LanguageUnionNonDet("<< rGen1.Name() 
	<< "," << rGen2.Name() << ")");
    
  // are state names enabled?
  bool stateNamesEnabled=rResGen.StateNamesEnabled();
    
  // use pointer pResGen to result rResGen; if rResGen is identical to
  // one of the parameters, allocate temporary object and copy back later
  vGenerator* pResGen = &rResGen;
  if(&rResGen== &rGen1 || &rResGen== &rGen2) {
    pResGen= rResGen.New();
  }

  // prepare result
  pResGen->Clear();

  // union of alphabets
  pResGen->InjectAlphabet(rGen1.Alphabet()+rGen2.Alphabet());
    
  // Maps from states of rGen1 and rGen2 to states of ResGen
  std::map<Idx,Idx> Gen1StatesMap;
  std::map<Idx,Idx> Gen2StatesMap;
    
  // "union" of states: insert states representing the states of rGen1 and rGen2
  StateSet::Iterator sit;
  for (sit = rGen1.StatesBegin(); sit != rGen1.StatesEnd(); ++sit) {
    if (stateNamesEnabled) {
      Gen1StatesMap[*sit] = pResGen->InsState(pResGen->UniqueStateName(rGen1.StateName(*sit)+"(1)"));
    }
    else {
      Gen1StatesMap[*sit] = pResGen->InsState();
    }
  }
  for (sit = rGen2.StatesBegin(); sit != rGen2.StatesEnd(); ++sit) {
    if (stateNamesEnabled) {
      Gen2StatesMap[*sit] = pResGen->InsState(pResGen->UniqueStateName(rGen2.StateName(*sit)+"(2)"));
    }
    else {
      Gen2StatesMap[*sit] = pResGen->InsState();
    }
  }
    
  // "union" of transition relations
  TransSet::Iterator tit;
  for (tit = rGen1.TransRelBegin(); tit != rGen1.TransRelEnd(); ++tit) {
    pResGen->SetTransition(Gen1StatesMap[tit->X1], tit->Ev, Gen1StatesMap[tit->X2]);
  }
  for (tit = rGen2.TransRelBegin(); tit != rGen2.TransRelEnd(); ++tit) {
    pResGen->SetTransition(Gen2StatesMap[tit->X1], tit->Ev, Gen2StatesMap[tit->X2]);
  }
    
  // "union" of init states
  for (sit = rGen1.InitStatesBegin(); sit != rGen1.InitStatesEnd(); ++sit) {
    pResGen->SetInitState(Gen1StatesMap[*sit]);
  }
  for (sit = rGen2.InitStatesBegin(); sit != rGen2.InitStatesEnd(); ++sit) {
    pResGen->SetInitState(Gen2StatesMap[*sit]);
  }
    
  // "union" of marked states
  for (sit = rGen1.MarkedStatesBegin(); sit != rGen1.MarkedStatesEnd(); ++sit) {
    pResGen->SetMarkedState(Gen1StatesMap[*sit]);
  }
  for (sit = rGen2.MarkedStatesBegin(); sit != rGen2.MarkedStatesEnd(); ++sit) {
    pResGen->SetMarkedState(Gen2StatesMap[*sit]);
  }
    
  // set name of result
  pResGen->Name(CollapsString("UnionNonDet("+rGen1.Name()+","+rGen2.Name()+")"));
    
  // if necessary, move pResGen to rResGen
  if(pResGen != &rResGen) {
    pResGen->Move(rResGen);
    delete pResGen;
  }
    
}

// LanguageUnion(rGen1, rGen2, rResGen)
void LanguageUnion(const vGenerator& rGen1, const vGenerator& rGen2, 
		   vGenerator& rResGen) {

  FD_DF("LanguageUnion("<< rGen1.Name() 
	<< "," << rGen2.Name() << ")");
   
  // fix name
  std::string name1 = rGen1.Name();
  std::string name2 = rGen2.Name();

  // avoid copy
  vGenerator* pTempGen = rResGen.New();
    
  // perform nondeterministic language union
  LanguageUnionNonDet(rGen1, rGen2, *pTempGen);
    
  // make deterministic
  Deterministic(*pTempGen, rResGen);

  // dispose temp
  delete pTempGen;

  // set name of result
  rResGen.Name(CollapsString("Union("+name1+","+name2+")"));

}

// LanguageIntersection(rGen1, rGen2, rResGen)
void LanguageIntersection(const vGenerator& rGen1, const vGenerator& rGen2, 
			  vGenerator& rResGen) {
  FD_DF("LanguageIntersection("<< rGen1.Name() 
	<< "," << rGen2.Name() << ")");
    
  // fix name
  std::string name1 = rGen1.Name();
  std::string name2 = rGen2.Name();

  // the product of rGen1 and rGen2 implements the language intersection
  Product(rGen1, rGen2, rResGen);
  rResGen.Name(CollapsString("Intersection("+name1+","+name2+")"));
    
}

// EmptyLanguageIntersection(rGen1, rGen2)
bool EmptyLanguageIntersection(const vGenerator& rGen1, const vGenerator& rGen2) {
  FD_DF("EmptyLanguageIntersection("<< rGen1.Name() 
	<< "," << rGen2.Name() << ")");
    
  // The product of rGen1 and rGen2 implements the language intersection
  // Note: rather use a function that does not explicitely compute the product 
  // but that returns false as soon as a common transtion is found.

  Generator ProductGen;
  Product(rGen1, rGen2, ProductGen);
  return IsEmptyLanguage(ProductGen);
}

// LanguageDisjoint(rGen1, rGen2)
bool LanguageDisjoint(const vGenerator& rGen1, const vGenerator& rGen2) {
  FD_DF("LanguageDisjoint("<< rGen1.Name() 
	<< "," << rGen2.Name() << ")");
  return EmptyLanguageIntersection(rGen1,rGen2);
}

// Automaton(rGen)
void Automaton(vGenerator& rGen, const EventSet& rAlphabet) {
  FD_DF("Automaton("<< rGen.Name() << "," << rAlphabet.Name() << ")");
    
  // extend rGen.Alphabet() by rAlphabet
  rGen.InjectAlphabet(rGen.Alphabet()+rAlphabet);
    
  // remove states that do not represent prefixes of marked strings
  rGen.Coaccessible();
    
  // the automaton for the empty language Lm={} contains only the dump state,
  // which, in this case, represents Sigma*
  if(IsEmptyLanguage(rGen)){
        
    // remember generator name
    std::string name; 
    name=rGen.Name();
        
    FullLanguage(rGen.Alphabet(),rGen);
    if (rGen.StateNamesEnabled()) {
      rGen.StateName(*(rGen.InitStatesBegin()),"dump");
    }
    // the dump state is not marked
    rGen.ClearMarkedStates();
        
    // set generator name
    rGen.Name(CollapsString("Automaton(" + name + "," + rAlphabet.Name() + ")"));
    return;
  }
    
  // for correct result, rGen has to be deterministic!
#ifdef FAUDES_CHECKED
  if ( !(rGen.IsDeterministic()) ) {
    FD_WARN("Automaton(): nondeterministic parameter " << rGen.Name() <<".");
  }
#endif 
    
  // prepare result
  rGen.Name(CollapsString("Automaton(" + rGen.Name() + "," + rAlphabet.Name() + ")"));
    
  // introduce dump state
  Idx dump;
  if (rGen.StateNamesEnabled()) {
    std::string dumpstr=rGen.UniqueStateName("dump");
    dump = rGen.InsState(dumpstr);
  } else {
    dump = rGen.InsState();
  }
	
  // introduce transitions to dumpstate
  StateSet::Iterator sit;
  EventSet::Iterator eit;
  bool dumpNotReached=true; // indicate that dumpstate is not reached
  for (sit = rGen.StatesBegin(); sit != rGen.StatesEnd(); ++sit) {        
    for (eit = rGen.Alphabet().Begin(); eit != rGen.Alphabet().End(); ++eit) {
      // If no transition is defined for this state and event, insert respective
      // transition to dump state (note that dump state itself is also treated here
      // and thus provided with selfloops)
      if (rGen.TransRelBegin(*sit, *eit) == rGen.TransRelEnd(*sit, *eit)) {
	rGen.SetTransition(*sit, *eit, dump);
	// indicate that dumstate was reached
	if(*sit!=dump) dumpNotReached=false;
      }
    }        
  }
    
  // if no transition was introduced (except for selfloops), remove dump state
  if(dumpNotReached) 
    rGen.DelState(dump);
}

// Automaton(rGen)
void Automaton(vGenerator& rGen) {
  FD_DF("Automaton("<< rGen.Name() << ")");
  std::string name=rGen.Name();
  Automaton(rGen,rGen.Alphabet());    
  rGen.Name(CollapsString("Automaton(" + name + ")"));
}

// LanguageComplement(rGen,rAlphabet)
void LanguageComplement(vGenerator& rGen, const EventSet& rAlphabet) {
  FD_DF("LanguageComplement("<< rGen.Name() << "," << rAlphabet.Name() << ")");

  // fix name
  std::string name = rGen.Name();  
  
  // convert to automaton (avoiding statename "dump")
  bool stateNamesEnabled=rGen.StateNamesEnabled(); 
  rGen.StateNamesEnabled(false); 
  Automaton(rGen,rAlphabet);
  rGen.StateNamesEnabled(stateNamesEnabled); 
    
  // invert marking
  rGen.InjectMarkedStates(rGen.States() - rGen.MarkedStates());

  // set name    
  rGen.Name(CollapsString("Complement(" + name + "," + rAlphabet.Name() + ")"));
}

// LanguageComplement(rGen)
void LanguageComplement(vGenerator& rGen) {
  FD_DF("LanguageComplement("<< rGen.Name() << ")");
  std::string name=rGen.Name();
  LanguageComplement(rGen,rGen.Alphabet());
  rGen.Name(CollapsString("Complement(" + name + ")"));
  return;    
}

// language complement, uniform api
void LanguageComplement(const vGenerator& rGen, vGenerator& rRes) {
  rRes=rGen;
  LanguageComplement(rRes);
}


// language complement, uniform api
void LanguageComplement(const vGenerator& rGen, const EventSet& rSigma, vGenerator& rRes) {
  rRes=rGen;
  LanguageComplement(rRes,rSigma);
}




//LanguageDifference(rGen1, rGen2, rResGen)
void LanguageDifference(
   const vGenerator& rGen1, 
   const vGenerator& rGen2,
   vGenerator& rResGen) {
         
     FD_DF("LanguageDifference("<< rGen1.Name() << "," << rGen2.Name() << ")");
     
    // use pointer pResGen to result rResGen; if rResGen is identical to
    // one of the parameters, allocate temporary object and copy back later 
    vGenerator* pResGen = &rResGen;
    if(&rResGen == &rGen1 || &rResGen== &rGen2) {
      pResGen = rResGen.New();
    }    
    
    // due to the use of LanguageComplement(), rGen2 has to be deterministic
    #ifdef FAUDES_CHECKED
    if(!(rGen2.IsDeterministic())){
      std::stringstream errstr;
      errstr << "Nondeterministic parameter " << rGen2.Name() << ".";
      throw Exception("LanguageDifference()", errstr.str(), 101);
    }
    #endif
          
    // prepare result
    pResGen->Clear();
     
    // calculate "Lm1-Lm2" by building the intersection of Lm1 with the complement of Lm2
        
    *pResGen=rGen2;
    // for correct result, complement has to be computed wrt the alphabet of Lm1!
    LanguageComplement(*pResGen,rGen1.Alphabet());
    LanguageIntersection(rGen1, *pResGen, *pResGen);
        
    pResGen->Name(CollapsString("LanguageDifference(" + rGen1.Name() + "," + rGen2.Name() + ")"));
  
    // if necessary, move pResGen to rResGen
    if(pResGen != &rResGen) {
      pResGen->Move(rResGen);
      delete pResGen;
    } 
    
    return;
}

// LanguageConcatenateNonDet(rGen1, rGen2, rResGen)
void LanguageConcatenateNonDet(const vGenerator& rGen1, const vGenerator& rGen2, 
			       vGenerator& rResGen) {
  FD_DF("LanguageConcatenateNonDet(" << rGen1.Name() << "," << rGen2.Name() << ")");
            
  // are state names enabled in result?
  bool stateNamesEnabled=rResGen.StateNamesEnabled();
    
  // use pointer pResGen to result rResGen; if rResGen is identical to
  // one of the parameters, allocate temporary object and copy back later
  vGenerator* pResGen = &rResGen;
  if(&rResGen== &rGen1 || &rResGen== &rGen2) {
    pResGen= rResGen.New();
  }

  // prepare result
  pResGen->Clear();
    
  // union of alphabets
  pResGen->InjectAlphabet(rGen1.Alphabet() + rGen2.Alphabet());

  // Maps from states of rGen1 and rGen2 to states of ResGen
  std::map<Idx,Idx> Gen1StatesMap;
  std::map<Idx,Idx> Gen2StatesMap;

  // helpers
  StateSet::Iterator sit;
  StateSet::Iterator sit1m, sit2i; // rGen1 marked states, rGen2 initial states
  TransSet::Iterator tit;
    
  // "union" of states: insert states representing the states of rGen1 and rGen2
  for (sit = rGen1.StatesBegin(); sit != rGen1.StatesEnd(); ++sit) {
    if (stateNamesEnabled) {
      Gen1StatesMap[*sit] = pResGen->InsState(pResGen->UniqueStateName(rGen1.StateName(*sit)+"(1)"));
    } else {
      Gen1StatesMap[*sit] = pResGen->InsState();
    }
  }
  for (sit = rGen2.StatesBegin(); sit != rGen2.StatesEnd(); ++sit) {
    if (stateNamesEnabled) {
      Gen2StatesMap[*sit] = pResGen->InsState(pResGen->UniqueStateName(rGen2.StateName(*sit)+"(2)"));
    } else {
      Gen2StatesMap[*sit] = pResGen->InsState();
    }
  }

  // test whether L2 includes the empty string
  bool concatenateEpsilon=false;
  for (sit = rGen2.InitStatesBegin(); sit != rGen2.InitStatesEnd(); ++sit) {
    if (rGen2.ExistsMarkedState(*sit)) {
      concatenateEpsilon=true;
    }
  }
    
  // set pResGen init states corresponding to rGen1 init states using Gen1StatesMap
  for (sit = rGen1.InitStatesBegin(); sit != rGen1.InitStatesEnd(); ++sit) {
    pResGen->SetInitState(Gen1StatesMap[*sit]);
  }
    
  // add rGen1 transitions
  for (tit = rGen1.TransRelBegin(); tit != rGen1.TransRelEnd(); ++tit) {
    pResGen->SetTransition(Gen1StatesMap[tit->X1], tit->Ev, Gen1StatesMap[tit->X2]);
  }
    
  // add rGen2 transitions
  for (tit = rGen2.TransRelBegin(); tit != rGen2.TransRelEnd(); ++tit) {
    pResGen->SetTransition(Gen2StatesMap[tit->X1], tit->Ev, Gen2StatesMap[tit->X2]);
  }
    
  // add rGen2 transitions starting at initial states at each state corresponding to a 
  // marked state of rGen1.
  for (sit2i = rGen2.InitStatesBegin(); sit2i != rGen2.InitStatesEnd(); ++sit2i) {
    for (tit = rGen2.TransRelBegin(*sit2i); tit != rGen2.TransRelEnd(*sit2i); ++tit) {
      // iterate over rGen1 mstates
        for (sit1m = rGen1.MarkedStatesBegin(); sit1m != rGen1.MarkedStatesEnd(); ++sit1m) {
            // If the rGen2 transition tit leads to an init state of rGen2, then the respective
            // pResGen transition has to lead to the current marked state of rGen1 (as selfloop)
            if(rGen2.ExistsInitState(tit->X2)){
                pResGen->SetTransition(Gen1StatesMap[*sit1m], tit->Ev, Gen1StatesMap[*sit1m]);
            }
            else {
                pResGen->SetTransition(Gen1StatesMap[*sit1m], tit->Ev, Gen2StatesMap[tit->X2]);
            }
            // note: in this state, there might already be a transition from rGen1 with the
            // same event. Hence, the transitions inserted above can cause nondeterminism in pResGen.
        }
    }
  }
    
  // set pResGen marked states corresponding to rGen2 marked states using Gen2StatesMap
  for (sit = rGen2.MarkedStatesBegin(); sit != rGen2.MarkedStatesEnd(); ++sit) {
    if(!(rGen2.ExistsInitState(*sit))) {
      pResGen->SetMarkedState(Gen2StatesMap[*sit]);
    }
  }

  // if epsilon is concatenated to Lm1, additionally set pResGen marked 
  // states corresponding to rGen1 marked states using Gen1StatesMap
  if(concatenateEpsilon){
    for (sit = rGen1.MarkedStatesBegin(); sit != rGen1.MarkedStatesEnd(); ++sit) {
      pResGen->SetMarkedState(Gen1StatesMap[*sit]);
    }
  }
    
  // remove blocking states as they provide no useful meaning.
  pResGen->Trim();
  pResGen->Name("ConcatenateNonDet("+rGen1.Name()+","+rGen2.Name()+")");
    
  // if necessary, move pResGen to rResGen
  if(pResGen != &rResGen) {
    pResGen->Move(rResGen);
    delete pResGen;
  }
    
}

// LanguageConcatenate(rGen1, rGen2, rResGen)
void LanguageConcatenate(const vGenerator& rGen1, const vGenerator& rGen2, 
			 vGenerator& rResGen) {

  FD_DF("LanguageConcatenate("<< rGen1.Name() 
	<< "," << rGen2.Name() << ")");
    
  // perform nondeterministic language concatenation
  LanguageConcatenateNonDet(rGen1, rGen2, rResGen);
    
  // make deterministic if necessary
  if(!(rResGen.IsDeterministic())){
    Deterministic(rResGen, rResGen);
  }

  // set name of result
  rResGen.Name("Concatenate("+rGen1.Name()+","+rGen2.Name()+")");
    
  return;    
}

// FullLanguage(rAlphabet, rResGen)
void FullLanguage(const EventSet& rAlphabet, vGenerator& rResGen) {
  FD_DF("FullLanguage("<< rAlphabet.Name() 
	<< "," << rResGen.Name() << ")");
	
  // prepare result
  rResGen.Clear();
	
  // helpers
  Idx state;
  EventSet::Iterator evit;
	
  // alphabet
  rResGen.InjectAlphabet(rAlphabet);
	
  // insert marked initial state
  if(rResGen.StateNamesEnabled()){
    state = rResGen.InsInitState("1");
  } else{
    state = rResGen.InsInitState();
  }
  rResGen.SetMarkedState(state);
	
  // create selfloop for each event
  for (evit = rAlphabet.Begin(); evit != rAlphabet.End(); ++evit) {
    rResGen.SetTransition(state, *evit, state);
  }
	
  // set name of result
  rResGen.Name("FullLanguage("+rAlphabet.Name()+")");
    
  return;
}

// AlphabetLanguage(rAlphabet, rResGen)
void AlphabetLanguage(const EventSet& rAlphabet, vGenerator& rResGen) {
  FD_DF("AlphabetLanguage("<< rAlphabet.Name() 
	<< "," << rResGen.Name() << ")");
	
  // prepare result
  rResGen.Clear();
    	
  // set name of result
  rResGen.Name("AlphabetLanguage("+rAlphabet.Name()+")");
    
  // if alphabet is empty, leave generator empty
  if(rAlphabet.Empty()){
    FD_WARN("AlphabetLanguage: empty alphabet.");
    return;
  }
    
  // helpers
  Idx istate, mstate;
  EventSet::Iterator evit;
	
  // alphabet
  rResGen.InjectAlphabet(rAlphabet);
	
  // insert one initial state and one marked state
  if(rResGen.StateNamesEnabled()){
    istate = rResGen.InsInitState("1");
    mstate = rResGen.InsMarkedState("2");
  }
  else{
    istate = rResGen.InsInitState();
    mstate = rResGen.InsMarkedState();
  }
	
  // for each event from rAlphabet, inserted transition leading from init state to marked state
  for (evit = rAlphabet.Begin(); evit != rAlphabet.End(); ++evit) {
    rResGen.SetTransition(istate, *evit, mstate);
  }
    
  return;
}

// EmptyStringLanguage(rAlphabet, rResGen)
void EmptyStringLanguage(const EventSet& rAlphabet, vGenerator& rResGen) {
  FD_DF("EmptyStringLanguage("<< rAlphabet.Name() 
	<< "," << rResGen.Name() << ")");
	
  // prepare result
  rResGen.Clear();
	
  // helpers
  Idx state;
	
  // alphabet
  rResGen.InjectAlphabet(rAlphabet);
	
  // insert marked initial state
  if(rResGen.StateNamesEnabled()){
    state = rResGen.InsInitState("1");
  }
  else{
    state = rResGen.InsInitState();
  }
  rResGen.SetMarkedState(state);
	
  // set name of result
  rResGen.Name("EmptyStringLanguage("+rAlphabet.Name()+")");
    
  return;
}

// EmptyLanguage(rAlphabet, rResGen)
void EmptyLanguage(const EventSet& rAlphabet, vGenerator& rResGen) {
  FD_DF("EmptyStringLanguage("<< rAlphabet.Name() 
	<< "," << rResGen.Name() << ")");
	
  // prepare result
  rResGen.Clear();
	
  // alphabet and initial state
  rResGen.InsInitState();
  rResGen.InjectAlphabet(rAlphabet);

  // set name of result
  rResGen.Name("EmptyLanguage("+rAlphabet.Name()+")");
    
  return;
}

// EmptyLanguage(rGen)
bool IsEmptyLanguage(const vGenerator& rGen) {
  // case a) check if set of marked states is empty
  if(rGen.MarkedStatesSize()==0) return true;
  // case b) check if no marked state is accessible (reachable)
  return (rGen.AccessibleSet()*rGen.MarkedStates()).Empty();
}

// LanguageInclusion(rGen1, rGen2)
bool LanguageInclusion(const vGenerator& rGen1, const vGenerator& rGen2) {
    
  FD_DF("LanguageInclusion("<< rGen1.Name() << "," << rGen2.Name() << ")"); 
    
  // check if there is no string in Lm1 that is not in Lm2, which means Lm1<=Lm2
  Generator NotrGen2=rGen2;
  // note: complement w.r.t. union of alphabets to ensure that elementwise 
  // inclusion of Lm1 in Lm2 is tested
  LanguageComplement(NotrGen2 , rGen1.Alphabet()+rGen2.Alphabet());
  return EmptyLanguageIntersection(rGen1,NotrGen2);        
}

// LanguageEquality(rGen1, rGen2)
bool LanguageEquality(const vGenerator& rGen1, const vGenerator& rGen2) {
    
  FD_DF("LanguageEquality("<< rGen1.Name() << "," << rGen2.Name() << ")");

  // Check for equality by testing mutual inclusion
  return LanguageInclusion(rGen1,rGen2) && LanguageInclusion(rGen2,rGen1);
}

// KleeneClosure(rGen, rResGen)
void KleeneClosure(vGenerator& rGen) {
   
  FD_DF("KleeneClosure("<< rGen.Name() << ")");

  // fix name 
  std::string name=CollapsString("KleeneClosure(" + rGen.Name() + ")");

  // The Kleene Closure of the empty set is the empty set
  if(IsEmptyLanguage(rGen)){
    rGen.Name(name);
    return;
  }

  // run nondet version    
  KleeneClosureNonDet(rGen);
  Deterministic(rGen, rGen);

  // set name
  rGen.Name(name);
}

// KleeneClosureNonDet(rGen)
void KleeneClosureNonDet(vGenerator& rGen) {
    
  FD_DF("KleeneClosureNonDet("<< rGen.Name()  << ")");
    
  // set name 
  rGen.Name(CollapsString("KleeneClosureNonDet("+ rGen.Name() + ")"));
    
  // The Kleene Closure of the empty set is the empty set
  if(IsEmptyLanguage(rGen)) return;
    
  // helpers
  TransSet::Iterator tit;
  StateSet::Iterator init = rGen.InitStatesBegin();
  TransSet TransToInsert;
    
  // for all transitions leading from a state x1 to a marked state: insert a transition
  // with the same event that leads to (one of) the initial state(s). This step makes
  // rGen nondeterministic.
  // first, find respective transitions in rGen and update TransToInsert with new
  // transitions to insert
  for (tit = rGen.TransRelBegin(); tit != rGen.TransRelEnd(); ++tit) {
	
    if(rGen.ExistsMarkedState(tit->X2)) {
      if( !(rGen.ExistsTransition(tit->X1, tit->Ev, *init)) ){
	TransToInsert.Insert(tit->X1, tit->Ev, *init);
      }
    }
        
  }
  // now, extend rGen by transitions of TransToInsert
  for (tit = TransToInsert.Begin(); tit != TransToInsert.End(); ++tit) {
    rGen.SetTransition(*tit);
  }
    
  // KleeneClosure of nonempty language always contains the empty string:
  // mark at least one initial state
  rGen.SetMarkedState(*init);
}

// PrefixClosure(rGen)
void PrefixClosure(vGenerator& rGen) {
    
  FD_DF("PrefixClosure("<< rGen.Name() << ")");

  // fix name
  std::string name=CollapsString("PrefixClosure("+ rGen.Name() + ")");
    
  // remove all states that do net represent prefixes of marked strings
  rGen.Coaccessible();
    
  // mark all remaining states
  rGen.InjectMarkedStates(rGen.States());
   
  // set name 
  rGen.Name(name);
    
}

// SelfLoop(rGen,rAlphabet)
void SelfLoop(vGenerator& rGen,const EventSet& rAlphabet) {
    
  FD_DF("SelfLoop(" << rGen.Name() << "," << rAlphabet.Name() << ")");

  // fix name
  std::string name = CollapsString("SelfLoop(" + rGen.Name() + "," + rAlphabet.Name() + ")");
  // extend alphabet of rGen
  rGen.InjectAlphabet(rGen.Alphabet()+rAlphabet);
    
  //helpers
  EventSet::Iterator evit,evbegin,evend;
  evbegin = rAlphabet.Begin();
  evend = rAlphabet.End();  
  StateSet::Iterator sit;
    
  // iterate over all states and insert selfloop for each event of rAlphabet
  for (sit = rGen.StatesBegin(); sit != rGen.StatesEnd(); ++sit) {
    for(evit = evbegin; evit != evend; ++evit){
      rGen.SetTransition(*sit, *evit, *sit);
    }
  }
    
  // set name
  rGen.Name(name);
}

// SelfLoopMarkedStates(rGen,rAlphabet)
void SelfLoopMarkedStates(vGenerator& rGen,const EventSet& rAlphabet) {
    
  FD_DF("SelfLoopMarkedStates(" << rGen.Name() << "," << rAlphabet.Name() << ")");

  // fix name
  std::string name = CollapsString("SelfLoopMarkedStates(" + rGen.Name() 
     + "," + rAlphabet.Name() + ")");

  // extend alphabet of rGen
  rGen.InjectAlphabet(rGen.Alphabet()+rAlphabet);
    
  //helpers
  EventSet::Iterator evit,evbegin,evend;
  evbegin = rAlphabet.Begin();
  evend = rAlphabet.End();  
  StateSet::Iterator sit;
    
  // iterate over all marked states and insert selfloop for each event of rAlphabet
  for (sit = rGen.MarkedStatesBegin(); sit != rGen.MarkedStatesEnd(); ++sit) {
    for(evit = evbegin; evit != evend; ++evit){
      rGen.SetTransition(*sit, *evit, *sit);
    }
  }
    
  // set name
  rGen.Name(name);
}

// SelfLoop(rGen,rAlphabet,rStates)
void SelfLoop(vGenerator& rGen,const EventSet& rAlphabet,const StateSet& rStates) {
    
  FD_DF("SelfLoop(" << rGen.Name() << "," << rAlphabet.Name() << "," << rStates.Name() << ")");

  // fix name
  std::string name = CollapsString("SelfLoop(" + rGen.Name() 
     + "," + rAlphabet.Name() + "," + rStates.Name() + ")");

  // exception: rStates must be states of rGen
#ifdef FAUDES_CHECKED
  if( !(rStates <= rGen.States()) ){
    std::stringstream errstr;
    errstr << "State set " << rStates.Name() << 
      " has to be included in state set of "<< rGen.Name() << ".";
    throw Exception("SelfLoop()", errstr.str(), 100);
  }
#endif
    
  // extend alphabet of rGen
  rGen.InjectAlphabet(rGen.Alphabet()+rAlphabet);
    
  //helpers
  EventSet::Iterator evit,evbegin,evend;
  evbegin = rAlphabet.Begin();
  evend = rAlphabet.End();    
  StateSet::Iterator sit;
    
  // iterate over all marked states and insert selfloop for each event of rAlphabet
  for (sit = rStates.Begin(); sit != rStates.End(); ++sit) {
    for(evit = evbegin; evit != evend; ++evit){
      rGen.SetTransition(*sit, *evit, *sit);
    }
  }
    
  // set name
  rGen.Name(name);
}

// rti wrapper
void LanguageDisjoint(const vGenerator& rGen1, const vGenerator& rGen2, bool& rRes) {
  rRes = LanguageDisjoint(rGen1,rGen2);
}

// rti wrapper
void LanguageInclusion(const vGenerator& rGen1, const vGenerator& rGen2, bool& rRes) {
  rRes = LanguageInclusion(rGen1,rGen2);
}

// rti wrapper
void LanguageEquality(const vGenerator& rGen1, const vGenerator& rGen2, bool& rRes) {
  rRes = LanguageEquality(rGen1,rGen2);
}

// rti wrapper
void IsEmptyLanguage(const vGenerator& rGen, bool& rRes) {
  rRes=IsEmptyLanguage(rGen);
}


} // namespace faudes

#undef Product //see define above for comment
