/** rti2code.cpp  Utility to generate registry initialisation code from rti files */

/* FAU Discrete Event Systems Library (libfaudes)

Copyright (C) 2009 Ruediger Berndt
Copyright (C) 2009 Thomas Moor

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */
	 

#include <string>
#include <iostream>
#include <fstream>
#include "corefaudes.h"


using namespace faudes;

int main(int argc, char *argv[]) {

  // simple command line
  if(argc != 3) {
    std::cerr << "rti2code: " << FDVersionString()  << std::endl;
    std::cerr << std::endl;
    std::cerr << "utility to generates c code from an rti-file to " << std::endl;
    std::cerr << "1) register faudes-types and -functions with the run-time interface," << std::endl;
    std::cerr << "2) extract c declarations for luabindings interface code." << std::endl;
    std::cerr << std::endl;
    std::cerr << "usage: rti2code <rti input file> <output basename>" << std::endl;
    exit(1);
  }

  // load registry 
  LoadRegistry(argv[1]);

  // code output streams
  std::ofstream rtiheader;
  rtiheader.open((std::string(argv[argc-1])+".h").c_str(), std::ios::out);
  std::ofstream rticode;
  rticode.open((std::string(argv[argc-1])+".cpp").c_str(), std::ios::out);
  std::ofstream luaheader;
  luaheader.open((std::string(argv[argc-1])+".i").c_str(), std::ios::out);

  // introduce myself
  rtiheader << "/* rti2code: autogenerated libFAUDES rti registration: "; 
  rtiheader << FDVersionString() << " "  << FDPluginsString() << " */" << std::endl << std::endl; 
  rticode   << "/* rti2code: autogenerated libFAUDES rti registration: "; 
  rticode   << FDVersionString() << " "  << FDPluginsString() << " */" << std::endl << std::endl; 
  luaheader << "/* rti2code: autogenerated libFAUDES luabindings declarations: "; 
  luaheader << FDVersionString() << " "  << FDPluginsString() << " */" << std::endl << std::endl; 

  // c function declaration: load types
  rtiheader << "namespace faudes {" << std::endl;
  rtiheader << "void LoadAutoregisteredTypes(void);" << std::endl;
  rtiheader << "} // namespace" << std::endl;

  // c function definition: load types
  rticode << "namespace faudes {" << std::endl;
  rticode << "/* Register faudes types */" << std::endl;
  rticode << "void LoadAutoregisteredTypes(void) {" << std::endl;

  // traverse type registry to figure faudes types
  TypeRegistry::Iterator  tit;
  for(tit=TypeRegistry::G()->Begin(); tit!=TypeRegistry::G()->End();tit++) {
    // get c type 
    std::string ctype=tit->second->CType();
    // bail out if no c type specified
    if(ctype=="") continue;
    // remove name space faudes
    size_t pos=ctype.find("faudes::");
    if(pos!=std::string::npos) 
      ctype=ctype.substr(std::string("faudes::").length());
    // report
    std::cout << "Generating code for: faudes type \"" << tit->second->Name() << "\"" << std::endl;
    // produce c code
    rticode << "  TypeRegistry::G()->Insert<" << ctype << ">(\"" << tit->second->Name() <<"\");";
    rticode << std::endl;
  }

  // c function definition: load types end
  rticode << "}" << std::endl;
  rticode << "} // namespace" << std::endl;


  // c function declaration: load functions
  rtiheader << "namespace faudes {" << std::endl;
  rtiheader << "void LoadAutoregisteredFunctions(void);" << std::endl;
  rtiheader << "} // namespace" << std::endl;

  // c function definition: load functions
  rticode << "namespace faudes {" << std::endl;
  rticode << "/* Register faudes functions */" << std::endl;
  rticode << "void LoadAutoregisteredFunctions(void) {" << std::endl;

  // c class definition: Function derivates 
  rtiheader << "namespace faudes {" << std::endl;

  // traverse function registry: define rti functions
  int fcnt=0;
  FunctionRegistry::Iterator  fit;
  for(fit=FunctionRegistry::G()->Begin(); fit!=FunctionRegistry::G()->End();fit++, fcnt++) {
    // current function definition
    const FunctionDefinition* fdef = fit->second;
    // get c type and faudes function name
    std::string ctype=fdef->CType();
    std::string fname = fdef->Name();  
    // bail out if no c type specified
    if(ctype=="") continue;
    // remove name space faudes
    size_t pos=ctype.find("faudes::");
    if(pos!=std::string::npos) 
      ctype=ctype.substr(std::string("faudes::").length());
    // bail out if no signature
    if(fdef->VariantsSize()==0) {
      std::cout << "Function registration: " << fname << ": no signatures" << std::endl;
      continue;
    }
    // interpret signatures: set up type array
    std::vector< std::vector<std::string> > cparams;
    std::vector< std::vector<Parameter::ParamAttr> > cattrib;
    cparams.resize(fdef->VariantsSize());
    cattrib.resize(fdef->VariantsSize());
    // loop all signatures
    for(int i=0; i<fdef->VariantsSize(); i++) {
      const Signature& sigi=fdef->Variant(i);
      for(int j=0; j<sigi.Size(); j++) {
        // retrieve faudes type and attrib
        std::string ftype=sigi.At(j).mTDName;
	Parameter::ParamAttr fattr=sigi.At(j).mAttr;
        // bail out on unknown faudestype
        if(!TypeRegistry::G()->Exists(ftype)) break;
        // get corresponding ctype
        std::string ctype=TypeRegistry::G()->Definition(ftype).CType();
        // bail out on unknown ctype
        if(ctype=="") break;
        // remove name space faudes
        size_t pos=ctype.find("faudes::");
        if(pos!=std::string::npos) 
          ctype=ctype.substr(std::string("faudes::").length());
        // param ok
        cparams.at(i).push_back(ctype);
        cattrib.at(i).push_back(fattr);
      } 
      // test for signature error
      if((int) cparams.at(i).size()!=sigi.Size()) {
        std::cout << "Function registration: " << fname << ": cannot interpret signature" << std::endl;
        cparams.resize(i);
        break;
      }
    }
    // report
    std::cout << "Generating code for: faudes function \"" << fdef->Name() << "\"" << std::endl;
    // produce c code: register all functions function
    std::string rtiname = std::string("Rti") + ToStringInteger(fcnt) + ctype;
    rticode << "  FunctionRegistry::G()->Insert<" << rtiname << ">(\"" << fname <<"\");" << std::endl;
    // produce c code: class declaration intro
    rtiheader << "/* Function class for c function " << ctype << "*/" << std::endl;
    rtiheader << "class " << rtiname << " : public Function { " << std::endl;
    rtiheader << "public:" << std::endl;
    rtiheader << rtiname << "(const FunctionDefinition* fdef) : Function(fdef) {};" << std::endl;
    rtiheader << "virtual Function* New(void) const { return new " << rtiname << "(pFuncDef); };" << std::endl;
    rtiheader << "protected:" << std::endl;
    // produce c code: function class: have typed param
    for(unsigned int i=0; i<cparams.size(); i++) 
      for(unsigned int j=0; j<cparams.at(i).size(); j++) 
        rtiheader << cparams.at(i).at(j) << "* " << "mP_" << i << "_" << j << ";" << std::endl;
    // produce c code: function class: do type check
    rtiheader << "virtual bool DoTypeCheck(int n) {" << std::endl;
    rtiheader << "  bool res=false;" << std::endl;
    rtiheader << "  switch(mVariantIndex) { "<< std::endl;
    for(unsigned int i=0; i<cparams.size(); i++) {
      rtiheader << "  case " << i << ": { // variant " << fdef->Variant(i).Name() << std::endl;
      rtiheader << "    switch(n) { "<< std::endl;
      for(unsigned int j=0; j<cparams.at(i).size(); j++) {
        rtiheader << "    case " << j << ": ";
        rtiheader << "    res=DoTypeCast<" << cparams.at(i).at(j) << ">(" << j << ", mP_" << i <<"_" << j << "); ";
        rtiheader << "break; "<< std::endl;
      }
      rtiheader << "    default: break; " << std::endl;
      rtiheader << "    } "<< std::endl;
      rtiheader << "    break; "<< std::endl;
      rtiheader << "  } "<< std::endl;
    }
    rtiheader << "  default: break; " << std::endl;
    rtiheader << "  } "<< std::endl;
    rtiheader << "  return res;" << std::endl;
    rtiheader << "};" << std::endl;
    // produce c code: function class: do execute
    rtiheader << "virtual void DoExecute(void) {" << std::endl;
    // produce c code: do execute: switch variant
    rtiheader << "  switch(mVariantIndex) { "<< std::endl;
    for(unsigned int i=0; i<cparams.size(); i++) {
      rtiheader << "  case " << i << ": { // variant " << fdef->Variant(i).Name() << std::endl;
      rtiheader << "    " << ctype <<"(";
      for(unsigned int j=0; j<cparams.at(i).size(); j++) {
        if(j!=0) rtiheader << " ,";
        // special case: integer
        if(cparams.at(i).at(j) == "Integer") {
          rtiheader << "*(mP_" << i << "_" << j << "->CReference())";
        } else 
        // special case: boolean
	  if(cparams.at(i).at(j) == "Boolean") {
          rtiheader << "*(mP_" << i << "_" << j << "->CReference())";
        } else
        // special case: integer
        if(cparams.at(i).at(j) == "String") {
          rtiheader << "*(mP_" << i << "_" << j << "->CReference())";
        } else
        // std case
        rtiheader << "*mP_" << i << "_" << j;
      }
      rtiheader << "); break; };" << std::endl;
    }
    // produce c code: switch variant; done
    rtiheader << "  default: break; " << std::endl;
    rtiheader << "  }; "<< std::endl;
    // produce c code: do execute: done
    rtiheader << "}; "<< std::endl;
    // produce c code: function class: done
    rtiheader << "};"  << std::endl;

    // produce c code: lua function definition(s)    
    luaheader << "/* faudes-function \"" << fname << "\" */" << std::endl;
    // figure my plugin to insert a conditional
    std::string plugin=fdef->PlugIn();
    luaheader << "#if " << "SwigModule == \"Swig" << plugin << "\"" << std::endl;
    // use ctype function name
    if(ctype!=fname)
      luaheader << "%rename(" << fname << ") " << ctype << ";" << std::endl;
    // generate per signature
    for(unsigned int i=0; i<cparams.size(); i++) {
      // create ctype function declaration
      luaheader << "void  " << ctype <<"(";
      for(unsigned int j=0; j<cparams.at(i).size(); j++) {
        if(j!=0) luaheader << ", ";
        // have const for +In+
	if(cattrib.at(i).at(j)==Parameter::In)
          luaheader << "const ";
        // special case: integer
        if(cparams.at(i).at(j) == "Integer") {
          luaheader << "long int&";
        } else 
        // special case: boolean
	if(cparams.at(i).at(j) == "Boolean") {
          luaheader << "bool&";
        } else
        // special case: string
        if(cparams.at(i).at(j) == "String") {
          luaheader << "std::string&";
        } else
        // std case ctype as refernce
	luaheader << cparams.at(i).at(j) << "&";
        // mark elementary outputs
        if(cparams.at(i).at(j) == "Boolean" || cparams.at(i).at(j) == "String" 
           || cparams.at(i).at(j) == "Integer")
	if(cattrib.at(i).at(j)==Parameter::Out)
          luaheader << " OUTPUT";
      }
      // end of function declaration
      luaheader << ");" << std::endl;
 
      // add help entry: build nice signature
      std::string luasig = " " + fname + "(";
      bool leftcomma = false;
      bool rightcomma = false;
      for(unsigned int j=0; j<cparams.at(i).size(); j++) {
        // special case: elementary output
        if(cparams.at(i).at(j) == "Boolean" || cparams.at(i).at(j) == "String" 
           || cparams.at(i).at(j) == "Integer")
	  if(cattrib.at(i).at(j)==Parameter::Out) {
            if(leftcomma) luasig = "," + luasig;
	    //            if(leftcomma) luasig = ", " + luasig; // need tab in help system?
	    luasig=cparams.at(i).at(j) + luasig;
            leftcomma=true;
	    continue;
	  }
        // std case
        if(rightcomma) luasig += ", ";
        const Signature& sigi=fdef->Variant(i);
	luasig += sigi.At(j).Str();
        rightcomma=true;
      }  
      luasig+=")";
      // add help entry: add with topic
      if(fdef->TextDoc()!=""){
        std::string topic= fdef->PlugIn();
        std::string key1=fdef->KeywordAt(1);
        if(topic=="CoreFaudes") {
          topic=fdef->KeywordAt(1);
          key1=fdef->KeywordAt(2);
        }
        if(topic.length()>0) topic.at(0)=toupper(topic.at(0));
        if(key1.length()>0) key1.at(0)=toupper(key1.at(0));
        luaheader << "SwigHelpEntry(\"" << topic <<"\", \"" << key1 << "\", \"" << luasig << "\");" << std::endl;
      }
    
    } // done: per signature;

    // and all signatures, incl conditional
    luaheader << "#endif " << std::endl;
    luaheader << std::endl;


  } // loop all functions

  // c class definition: function class: all such done
  rtiheader << "} // namespace" << std::endl;

  // c class definition: register function prototypes: done
  rticode << "}" << std::endl;
  rticode << "} // namespace" << std::endl;

  return(0);
}
