/** @file rtitypes.cpp Runtime interface, faudes types */

/* FAU Discrete Event Systems Library (libfaudes)

Copyright (C) 2009 Ruediger Berndt
Copyright (C) 2009 Thomas Moor

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#include "rtitypes.h"

namespace faudes{

/*
********************************************************************
********************************************************************
********************************************************************

implemantation of faudes Type

********************************************************************
********************************************************************
********************************************************************
*/

// static members: ref to the only one instance
const std::string Type::mName("TypeStdName");

// constructor
Type::Type(void)  {}

// copy constructor
Type::Type(const Type& rType)  {(void) rType;}

// destructor
Type::~Type(void) {}

// pointer constructor
Type* Type::New(void) const { 
  return new Type(); 
}

// pointer copy constructor
Type* Type::Copy(void) const { 
  return new Type(*this); 
}

// cast
const Type* Type::Cast(const Type* pOther) const {
  return pOther;
}

// Clear()
void Type::Clear(void) {
  FD_DC("Type::Clear(): not re-implemented in " << typeid(*this).name());
}

// assign
Type& Type::Assign(const Type& rSource) {
  FD_DC("Type(" << this << ")::Assign(" << &rSource << ")");
  Clear();
  return *this;
}

// equality
bool Type::Equal(const Type& rOther) const { 
  return true;
}   

// equality
bool Type::operator==(const Type& rOther) const { 
  return true;
}

// equality
bool Type::operator!=(const Type& rOther) const { 
  return false;
}

// assign
Type& Type::operator=(const Type& rSource) {
  FD_DC("Type(" << this << ")::AssignementOperator(" << &rSource << ")");
  Clear();
  return *this;
}

// assign
Type& Type::DoAssign(const Type& rSource) {
  FD_DC("Type(" << this << ")::DoAssign(" << &rSource << ")");
  return *this;
}

// equality
bool Type::DoEqual(const Type& rOther) const {
  FD_DC("Type(" << this << ")::DoEqual(" << &rOther << ")");
  return true;
}

// Name(rName)
void Type::Name(const std::string& rName) {
}

// Name()
const std::string& Type::Name(void) const {
  return mName;
}

// Write(context)
void Type::Write(const Type* pContext) const {
  TokenWriter tw(TokenWriter::Stdout);
  DoWrite(tw,"",pContext);
}

// Write(rFileName, label, context, openmode)
void Type::Write(const std::string& rFileName, const std::string& rLabel, 
		 const Type* pContext, std::ios::openmode openmode) const {
  try {
    TokenWriter tw(rFileName, openmode);
    DoWrite(tw, rLabel, pContext);
  }
  catch (std::ios::failure&) {
    std::stringstream errstr;
    errstr << "Exception opening/writing file \"" << rFileName << "\"";
    throw Exception("Type::Write", errstr.str(), 2);
  }
}

// Write(rFileName, openmode)
void Type::Write(const std::string& rFileName, std::ios::openmode openmode) const {
  Write(rFileName,"",0,openmode);
}

// Write(tw,label,context)
void Type::Write(TokenWriter& rTw, const std::string& rLabel, const Type* pContext) const {
  DoWrite(rTw,rLabel,pContext);
}

// ToString(label, context)
std::string Type::ToString(const std::string& rLabel, const Type* pContext) const {
  TokenWriter tw(TokenWriter::String);
  tw.Endl(false);
  DoWrite(tw,rLabel,pContext);
  return tw.Str();
}

// ToSText()
std::string Type::ToSText(void) const {
  TokenWriter tw(TokenWriter::String);
  tw.Endl(true);
  DoSWrite(tw);
  return tw.Str();
}

// ToString(label, context)
std::string Type::ToText(const std::string& rLabel, const Type* pContext) const {
  TokenWriter tw(TokenWriter::String);
  tw.Endl(true);
  DoWrite(tw,rLabel,pContext);
  return tw.Str();
}

// DWrite(context)
void Type::DWrite(const Type* pContext) const {
  TokenWriter tw(TokenWriter::Stdout);
  DoDWrite(tw,"",pContext);
}

// DWrite(rFileName, label, context, openmode)
void Type::DWrite(const std::string& rFileName, const std::string& rLabel, 
		  const Type* pContext, std::ios::openmode openmode) const {
  try {
    TokenWriter tw(rFileName, openmode);
    DoDWrite(tw, rLabel, pContext);
  }
  catch (std::ios::failure&) {
    std::stringstream errstr;
    errstr << "Exception opening/writing file \"" << rFileName << "\"";
    throw Exception("Type::DWrite", errstr.str(), 2);
  }
}

// DWrite(tw,label,context)
void Type::DWrite(TokenWriter& rTw, const std::string& rLabel, const Type* pContext) const {
  DoDWrite(rTw,rLabel,pContext);
}

// SWrite()
void Type::SWrite(void) const {
  TokenWriter tw(TokenWriter::Stdout);
  DoSWrite(tw);
}

// SWrite(tw)
void Type::SWrite(TokenWriter& rTw) const {
  DoSWrite(rTw);
}

// Read(rFilename, rLabel,context)
void Type::Read(const std::string& rFilename, const std::string& rLabel, const Type* pContext) {
  Clear();
  TokenReader tr(rFilename);
  DoRead(tr,rLabel,pContext);
}

// Read(rTr, rLabel, context)
void Type::Read(TokenReader& rTr, const std::string& rLabel, const Type* pContext) {
  Clear();
  DoRead(rTr,rLabel,pContext);
}

// FromString(rString, rLabel, context)
void Type::FromString(const std::string& rString, const std::string& rLabel, const Type* pContext) {
  Clear();
  TokenReader tr(TokenReader::String,rString);
  DoRead(tr,rLabel,pContext);
}
  
//DoWrite(rTr,rLabel,pContext)
void Type::DoWrite(TokenWriter& rTw, const std::string& rLabel, const Type* pContext) const {
  (void) rTw; (void) rLabel; (void) pContext;
  FD_DC("Type::DoWrite(): not re-implemented in " << typeid(*this).name());
}

//DoDWrite(rTr,rLabel,pContext)
void Type::DoDWrite(TokenWriter& rTw, const std::string& rLabel, const Type* pContext) const {
  FD_DC("Type::DoDWrite(): not re-implemented in " << typeid(*this).name() << ", using DoDWrite instead");
  DoWrite(rTw,rLabel,pContext);
}

//DoSWrite(rTr,rLabel,pContext)
void Type::DoSWrite(TokenWriter& rTw) const {
  FD_DC("Type::DoSWrite(): not re-implemented in " << typeid(*this).name());
  rTw.WriteComment("");
  rTw.WriteComment(" Statistics for " + Name());
  rTw.WriteComment("");
}

//DoRead(rTr,rLabel,pContext)
void Type::DoRead(TokenReader& rTr, const std::string& rLabel, const Type* pContext) {
  (void) rLabel; (void) pContext; (void) rTr;
  FD_DC("Type::DoRead(): not re-implemented in " << typeid(*this).name());
}



/*
********************************************************************
********************************************************************
********************************************************************

Implementation of class Documentation

********************************************************************
********************************************************************
********************************************************************
*/

// construct
Documentation::Documentation(){
}

// clear all
void Documentation::Clear(void){
  FD_DRTI("Documentation::Clear()");
  mName="";
  mPlugIn="CoreFaudes";
  mHtmlDoc="";
  mTextDoc="";
  mKeywords="";
}

//  read access members
const std::string& Documentation::Name(void) const{ return mName; }
const std::string& Documentation::PlugIn(void) const{ return mPlugIn; }
const std::string& Documentation::CType(void) const{ return mCType; }
const std::string& Documentation::TextDoc() const{ return mTextDoc; }
const std::string& Documentation::HtmlDoc() const{ return mHtmlDoc; }
const std::string& Documentation::Keywords() const{ return mKeywords; }

//  write access members
void Documentation::Name(const std::string& name){mName = name; }
void Documentation::PlugIn(const std::string& plugin){mPlugIn = plugin; }
void Documentation::CType(const std::string& name){mCType = name; }
void Documentation::TextDoc(const std::string& textdoc){mTextDoc = textdoc;}
void Documentation::HtmlDoc(const std::string& fname){mHtmlDoc = fname;}

//  write access keyword
void Documentation::AddKeyword(const std::string& rKeyword){
  if(mKeywords.empty()){
    mKeywords = rKeyword;
  } else{
    mKeywords.append(1, mDelim);
    mKeywords.append(rKeyword);
  }
}

// regex match keywords (todo)
std::string Documentation::MatchKeyword(const std::string& rPattern) const{
  FD_DRTI("Documentation::MatchKeyword(" << rPattern << ")");
  // be case insensitive (inefficient)
  std::string keys = mKeywords;
  std::transform(keys.begin(), keys.end(), keys.begin(), tolower);
  std::string match = rPattern;
  std::transform(match.begin(), match.end(), match.begin(), tolower);
  // find match
  std::string res="";
  std::size_t posk=keys.find(match);
  if(posk==std::string::npos) return res;
  // find delimiters 
  std::size_t posa=0;
  std::size_t posb=keys.length();
  while(true) {
    posb=keys.find(mDelim,posa);
    if(posb==std::string::npos) posb=mKeywords.length();
    if(posb>posk) break;
    posa=posb+1;
  };
  // return original case match
  FD_DRTI("Documentation::MatchKeyword(" << rPattern << "): " << mKeywords.substr(posa,posb-posa));
  return mKeywords.substr(posa,posb-posa);
}


// positional keyword
std::string Documentation::KeywordAt(int pos) const{
  FD_DRTI("Documentation::KeywordAt(" << pos << ")");
  // default result
  std::string res="";
  // find delimiter
  std::size_t posa=0;
  std::size_t posb=mKeywords.length();
  while(pos>0) {
    if(posa>=mKeywords.length()) return res;
    posb=mKeywords.find(mDelim,posa);
    if(posb==std::string::npos) return res;
    posa=posb+1;
    pos--;
  };
  // default result
  res=mKeywords.substr(posa);
  // find next delimiter
  posb=res.find(mDelim);
  if(posb==std::string::npos) return res;
  res=res.substr(0,posb);
  return res;
}

// keyword
int Documentation::KeywordsSize(void) const{
  int cnt=0;
  // find delimiter
  std::size_t pos=0;
  while(pos<mKeywords.length()) {
    cnt++;
    pos=mKeywords.find(mDelim,pos);
    if(pos==std::string::npos) break;
    pos++;
  };
  return cnt;
}


// token read
void Documentation::DoRead(TokenReader& rTr,  const std::string& rLabel, const Type* pContext) {
  FD_DRTI("Documentation::DoRead(): file " << rTr.FileName());
  // ignore label and context
  (void) rLabel;
  (void) pContext;
  Token token;
  // my faudes entety (eg type name incl plugin)
  std::string ident= rTr.ReadString();
  // separate name
  size_t pos=ident.find("::");
  if(pos!=std::string::npos) mName=ident.substr(pos+2);
  else mName=ident;
  // separate plugin
  if(pos!=std::string::npos) mPlugIn=ident.substr(0,pos);
  else mPlugIn = "CoreFaudes";
  // corresponding c type
  mCType="";
  rTr.Peek(token);
  if(token.Type()==Token::Option)
    mCType = rTr.ReadOption();
  // loop other known entries
  while(true) {
    rTr.Peek(token);
    // do the read: text
    if(token.Type()==Token::Begin)  
    if(token.StringValue()=="TextDoc") {
      rTr.ReadBegin("TextDoc");
      mTextDoc = rTr.ReadString();
      rTr.ReadEnd("TextDoc");
      continue;
    }
    // do the read: html
    if(token.Type()==Token::Begin)  
    if(token.StringValue()=="HtmlDoc") {
      rTr.ReadBegin("HtmlDoc");
      mHtmlDoc = rTr.ReadString();
      rTr.ReadEnd("HtmlDoc");
      continue;
    }
    // do the read: keys
    if(token.Type()==Token::Begin)  
    if(token.StringValue()=="Keywords") {
      rTr.ReadBegin("Keywords");
      while(!rTr.Eos("Keywords")) 
        AddKeyword(rTr.ReadString());
      rTr.ReadEnd("Keywords");
      continue;
    }
    // unknown -> derived class take over
    break;
  }
}


// token write
void Documentation::DoWrite(TokenWriter& rTw, const std::string& rLabel, const Type* pContext) const {
  // ignore label and context
  (void) rLabel;
  (void) pContext;
  // write name
  rTw << mPlugIn + "::" + mName ;
  // write c type
  if(mCType!="") rTw.WriteOption(mCType);
  // write text doc
  rTw.WriteBegin("TextDoc");
  rTw << mTextDoc;
  rTw.WriteEnd("TextDoc");
  // write html doc
  rTw.WriteBegin("HtmlDoc");
  rTw << mHtmlDoc;
  rTw.WriteEnd("HtmlDoc");
  // write keys
  rTw.WriteBegin("Keywords");
  rTw << mKeywords;
  rTw.WriteEnd("Keywords");
}



/*
********************************************************************
********************************************************************
********************************************************************

Implementation of class TypeDefinition

********************************************************************
********************************************************************
********************************************************************
*/


// typedefinition constructor function
TypeDefinition* TypeDefinition::Constructor(Type* pProto, const std::string& rTypeName){
  FD_DRTI("TypeDefinition::Construct(" << typeid(pProto).name() << ", " << rTypeName << ")");
  TypeDefinition* td = new TypeDefinition();
  td->Prototype(pProto);
  std::string name=rTypeName;
  if(name=="") name=typeid(pProto).name();
  td->Name(name);
  return(td);
}

// merge with docu from token io
void TypeDefinition::MergeDocumentationBody(TokenReader& rTr) {
  FD_DRTI("TypeDefinition::MergeDocumentationBody(): file " << rTr.FileName() << " for " << mName);
  // peek type ident
  Token token;
  rTr.Peek(token);
  // separate name
  std::string name = token.StringValue();
  size_t pos=name.find("::");
  if(pos!=std::string::npos) name=name.substr(pos+2);
  // match?
  if(mName!=name && mName!="") {
    std::stringstream errstr;
    errstr << "Documentation mismatch in file \"" << rTr.FileName() << "\" : " << mName << "!=" << name;
    throw Exception("TypeDefinition::MergeDocumentationBody", errstr.str(), 48);
  };
  // call base
  Documentation::DoRead(rTr);  
}

// token io
void TypeDefinition::DoRead(TokenReader& rTr,  const std::string& rLabel, const Type* pContext) {
  FD_DRTI("TypeDefinition::DoRead(): file " << rTr.FileName());
  // ignore
  (void) rLabel;
  (void) pContext;
  // doit
  rTr.ReadBegin("TypeDefinition");
  // peek and set name
  Token token;
  rTr.Peek(token);
  // extract name 
  std::string mName=token.StringValue();
  size_t pos=mName.find("::");
  if(pos!=std::string::npos) mName=mName.substr(pos+2);
  // do read
  MergeDocumentationBody(rTr);
  // no extras .. done
  rTr.ReadEnd("TypeDefinition");
}

// token io
void TypeDefinition::DoWrite(TokenWriter& rTw,  const std::string& rLabel, const Type* pContext) const {
  // ignore
  (void) rLabel;
  (void) pContext;
  // doit
  rTw.WriteBegin("TypeDefinition");
  // call base
  Documentation::DoWrite(rTw); 
  // no extras .. done
  rTw.WriteEnd("TypeDefitinion");
}

// access prototype
const Type* TypeDefinition::Prototype(void) const{
  return(mpType);
}

// set prototype
void TypeDefinition::Prototype(faudes::Type* pType){
  if(mpType) delete mpType;
  mpType = pType;
}

// construct new object with faudes type
Type* TypeDefinition::NewObject() const{
  FD_DRTI("TypeDefinition::NewObject()");
  // bail out 
  if(!mpType) return NULL;
  // use prototype
  return mpType->New();
}


// clear (all except prototype)
void TypeDefinition::Clear(){
  FD_DRTI("TypeDefinition::Clear()");
  // call base
  Documentation::Clear();
}



/*
********************************************************************
********************************************************************
********************************************************************

Implementation of class Integer

********************************************************************
********************************************************************
********************************************************************
*/


// public:
Integer::Integer() {
  CValue(0);
}

Integer::Integer(SignedIdx val) {
  CValue(val);
}

Type* Integer::New() const{
  return(new Integer());
}

Type* Integer::Copy() const{
  return(new Integer(mCInteger));
}

void Integer::CValue(SignedIdx val){
  mCInteger = val;
}

SignedIdx Integer::CValue() const{
  return(mCInteger);
}

SignedIdx* Integer::CReference() {
  return &mCInteger;
}

void Integer::DoWrite(TokenWriter& rTw, const std::string& rLabel, const Type* pContext) const{
  (void) pContext;
  std::string label=rLabel;
  if(label=="") label="Integer";
  FD_DRTI("Integer::DoWrite(): file " << rTw.FileName() << " section " << label);
  rTw.WriteBegin(label);
  rTw.WriteFloat(mCInteger);
  rTw.WriteEnd(label);
}

void Integer::DoRead(TokenReader& rTr,  const std::string& rLabel, const Type* pContext) {
  (void) pContext;
  std::string label = rLabel;
  if(label == "") label = "Integer";
  FD_DRTI("Integer()::DoRead(): file " << rTr.FileName() << "  section " << label);
  rTr.SeekBegin(label);
  mCInteger = (SignedIdx)rTr.ReadFloat();
  rTr.SeekEnd(label);
  FD_DRTI("Integer::DoRead(): done");
}


// integer sum, uniform rti api
void IntegerSum(const Integer& arg1, const Integer& arg2, Integer& res) {
  res=arg1+arg2;
}


/*
********************************************************************
********************************************************************
********************************************************************

Implementation of class String

********************************************************************
********************************************************************
********************************************************************
*/


// constructor
String::String() {
  CValue("");
}

// constructor
String::String(std::string val) {
  CValue(val);
}

// factory constructor
Type* String::New() const{
  return(new String());
}

// factory constructor
Type* String::Copy() const{
  return(new String(mCString));
}

// c value
std::string String::CValue() const{
  return(mCString);
}

// c value
void String::CValue(std::string s){
  mCString = s;
}

// c ref
std::string* String::CReference() {
  return &mCString;
}

// token io
void String::DoWrite(TokenWriter& rTw, const std::string& rLabel, const Type* pContext) const{
  (void)pContext;
  std::string label = rLabel;
  if(label == "") label = "String";
  FD_DRTI("String::DoWrite(): file " << rTw.FileName() << " section " << label);
  rTw.WriteBegin(label);
  rTw.WriteString(mCString);
  rTw.WriteEnd(label);
}

// token io
void String::DoRead(TokenReader& rTr,  const std::string& rLabel, const Type* pContext) {
  (void) pContext;
  std::string label = rLabel;
  if(label == "") label = "String";
  FD_DRTI("String::DoRead(): file " << rTr.FileName() << "  section " << label);
  rTr.SeekBegin(label);
  mCString = rTr.ReadString();
  rTr.SeekEnd(label);
  FD_DRTI("String::DoRead(): done");
}


/*
********************************************************************
********************************************************************
********************************************************************

Implementation of class Boolean

********************************************************************
********************************************************************
********************************************************************
*/


// constructor
Boolean::Boolean() {
  CValue(true);
}

// constructor
Boolean::Boolean(bool val) {
  CValue(val);
}

// factory constructor
Type* Boolean::New() const{
  return(new Boolean());
}

// factory constructor
Type* Boolean::Copy() const{
  return(new Boolean(mCBool));
}


// cvaliu
void Boolean::CValue(bool val){
  mCBool = val;
}

// cvalue 
bool Boolean::CValue() const{
  return(mCBool);
}

// c ref
bool* Boolean::CReference() {
  return &mCBool;
}

// token io
void Boolean::DoWrite(TokenWriter& rTw, const std::string& rLabel, const Type* pContext) const{
  (void)pContext;
  std::string label = rLabel;
  if(label == "") label = "Boolean";
  FD_DRTI("Boolean::DoWrite(): file " << rTw.FileName() << " section " << label);
  rTw.WriteBegin(label);
  if(mCBool) rTw.WriteString("true");
  else rTw.WriteString("false");
  rTw.WriteEnd(label);
}


// token io
void Boolean::DoRead(TokenReader& rTr,  const std::string& rLabel, const Type* pContext) {
  (void) pContext;
  std::string label = rLabel;
  if(label == "") label = "Boolean";
  FD_DRTI("Boolean::DoRead(): file " << rTr.FileName() << "  section " << label);
  rTr.SeekBegin(label);
  std::string value = rTr.ReadString();
  std::transform(value.begin(), value.end(), value.begin(), tolower);
  if(value=="true") mCBool=true;
  else if(value=="false") mCBool=false;
  else {
    std::stringstream err;
    err << "Expected true or false: " << rTr.FileLine();
    throw Exception("Boolean::DoRead()", err.str(), 52);
  }
  rTr.SeekEnd(label);
  FD_DRTI("Boolean::DoRead(): done");
}



} //namspace

