/** @file tokenreader.cpp @brief Class TokenReader */

/* FAU Discrete Event Systems Library (libfaudes)

Copyright (C) 2006  Bernd Opitz
Copyright (C) 2006  Thomas Moor
Exclusive copyright is granted to Klaus Schmidt

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */




#include "tokenreader.h"

namespace faudes {

// TokenReader(mode,instring)
TokenReader::TokenReader(Mode mode, const std::string& rInString)
  : mMode(mode), mpStream(NULL), mFileName("") 
{
  switch(mode) {
  case String: 
    // use mSStream
    FD_DV("TokenReader::Tokenreader(String, ...): " << rInString);
    mpSStream= new std::istringstream(rInString, std::istringstream::in | std::istringstream::binary);
    mpStream= mpSStream;
    Rewind();
    break;
  case File: 
    // set up mFStream
    FD_DV("TokenReader::Tokenreader(File, \"" << rInString <<"\")");
    mFStream.exceptions(std::ios::badbit|std::ios::failbit);
    try{
      mFStream.open(rInString.c_str(), std::ios::in | std::ios::binary); 
    } 
    catch (std::ios::failure&) {
      std::stringstream errstr;
      errstr << "Exception opening/reading file \""<< rInString << "\"";
      throw Exception("TokenReader::TokenReader", errstr.str(), 1);
    }
    mFileName=rInString;
    mpStream=&mFStream;
    Rewind();
    break;
  default:
    std::stringstream errstr;
    errstr << "Invalid Mode / Not implemented";
    throw Exception("TokenReader::TokenReader(mode,instring)", errstr.str(), 1);
  }
}


// TokenReader(rFilename)
TokenReader::TokenReader(const std::string& rFilename) 
  : mMode(File), mpStream(NULL), mFileName(rFilename) 
{
  // set up mFStream
  FD_DV("TokenReader::Tokenreader(File, \"" << rFilename <<"\")");
  mFStream.exceptions(std::ios::badbit|std::ios::failbit);
  try{
    mFStream.open(rFilename.c_str(), std::ios::in | std::ios::binary); 
  } 
  catch (std::ios::failure&) {
    std::stringstream errstr;
    errstr << "Exception opening/reading file \""<< rFilename << "\"";
    throw Exception("TokenReader::TokenReader", errstr.str(), 1);
  }
  mFileName=rFilename;
  mpStream=&mFStream;
  Rewind();
}


//  destruct
TokenReader::~TokenReader(void) {
  if(mMode==String) delete mpSStream;
}

// Rewind()
void TokenReader::Rewind(void) {
  FD_DV("TokenReader::Rewind: \"" << mFileName <<"\"");
  try {
    mpStream->clear();
    mpStream->seekg(0);
    mHasPeekToken=false;
    mLevel=0;
    mLevelPos.clear();
    mLevelPos.push_back(0);
    mLevelLine.clear();
    mLevelLine.push_back(1);
    mSeekLevel.clear();
    mLineCount=1;
    mFilePos=0;
  }
  catch (std::ios::failure&) {
    std::stringstream errstr;
    errstr << "Exception opening/reading file in "<< FileLine();
    throw Exception("TokenReader::Rewind", errstr.str(), 1);
  }
}


// FileName()
std::string TokenReader::FileName(void) const {
  return mFileName;
}

// Peek(token)
bool TokenReader::Peek(Token& token) {
  if(mHasPeekToken) { 
    token=mPeekToken;
    return true;
  }
  try{
    mLineCount += token.Read(mpStream);
  } 
  catch (std::ios::failure&) {
    std::stringstream errstr;
    errstr << "Exception opening/reading file in "<< FileLine();
    throw Exception("TokenReader::Peek", errstr.str(), 1);
  }
  mPeekToken=token;
  if (mPeekToken.Type() == Token::None) {
    return false;
  } 
  mHasPeekToken=true;
  return true;
}

// Get(token)
bool TokenReader::Get(Token& token) {
  bool res;
  // get token either from peek buffer or stream
  if(mHasPeekToken) {   
    res=true;
    token=mPeekToken;
    mHasPeekToken=false;
  } else {
    res=Peek(token);
    mHasPeekToken=false;
  }    
  // track state (level of nested sections, filepos etc)
  if(res) {
    mFilePos=mpStream->tellg();
    if (token.Type() == Token::Begin) {
      mLevel++;
      mLevelPos.push_back(mFilePos);
      mLevelLine.push_back(mLineCount);
    }
    if (token.Type() == Token::End) {
      mLevel--;
      mLevelPos.pop_back();
      mLevelLine.pop_back();
    }
  }
  return res;
}

// SeekBegin(label)
void TokenReader::SeekBegin(const std::string& rLabel) {
  // search for begin at any descending level, no rewind
  // if successful, must be followed by SeekEnd
  FD_DV("TokenReader::SeekBegin: " << rLabel << " at " << FileLine() << " level " << mLevel);
  mSeekLevel.push_back(mLevel);
  int level=mLevel;
  long int startpos=mFilePos;
  int startline=mLineCount;
  Token token;
  for (;;) {
    // exception: did not get a token at all (incl. eof)
    if(!Peek(token)) {
      Rewind();
      std::stringstream errstr;
      errstr << "Section \"" << rLabel << "\" expected at " << FileLine() << " no more tokens";
      throw Exception("TokenReader::SeekBegin", errstr.str(), 51);
    }
    // exception: current section ends
    if ((token.Type() == Token::End) && (mLevel == level)) {
      mpStream->seekg(startpos);
      mLineCount=startline;
      mSeekLevel.pop_back();
      mHasPeekToken=false;
      std::stringstream errstr;
      errstr << "Section \"" << rLabel << "\" expected at " << FileLine() 
	     << "current section ended unexpected. Found: " << token.StringValue() << " Type " << token.Type();
      throw Exception("TokenReader::SeekBegin", errstr.str(), 51);
    }
    Get(token);
    // success: found begin section
    if ((token.Type() == Token::Begin) && (token.StringValue() == rLabel))
      break;
  }
}
 
// SeekEnd(label)
void TokenReader::SeekEnd(const std::string& rLabel) {
  // search for end at current level ...
  ReadEnd(rLabel); // throws id 51
  // ... then climb up to undo recent seek
  if(mSeekLevel.size()==0) {
    std::stringstream errstr;
    errstr << "unmatched call";
    throw Exception("TokenReader::SeekEnd", errstr.str(), 52);
  }
  int level=mSeekLevel.back();
  mSeekLevel.pop_back();
  FD_DV("TokenReader::SeekEnd: " << rLabel << " at " << FileLine() << " level " << mLevel << " for " << level);
  Token token;
  for (;;) {
    // success 
    if(mLevel==level) {
      break;
    }
    // exception: did not get a token at all
    if(!Get(token)) {
      std::stringstream errstr;
      errstr << "could not find level " << level << " " << FileLine();
      throw Exception("TokenReader::SeekEnd", errstr.str(), 52);
    }
  }
}

// ReadBegin(label)
void TokenReader::ReadBegin(const std::string& rLabel) {
  FD_DV("looking for Section \"" << rLabel << "\"");
  try {
    int level=mLevel;
    bool firstgo=true;
    long int startpos=mFilePos;
    FD_DV("section level " << level << " current pos " << startpos << " begin of section " << mLevelPos[level]);
    Token token;
    // search for begin at current level
    for (;;) {
      // exception: did not get a token at all (incl eof)
      if(!Peek(token)) {
	std::stringstream errstr;
	errstr << "Section \"" << rLabel << "\" expected at " << FileLine() << ", no token at all";
	throw Exception("TokenReader::ReadBegin Peek", errstr.str(), 51);
      }
      // success: found begin section
      if ((token.Type() == Token::Begin) && (token.StringValue() == rLabel) && (mLevel==level)) {
	Get(token);
	break;
      }
      // rewind once when current section ends
      if ((token.Type() == Token::End) && (mLevel == level) && firstgo) {
	mpStream->seekg(mLevelPos[level]);
	mLineCount=mLevelLine[level];                       
	firstgo=false;
	mHasPeekToken=false;
	continue;
      }
      // exception: did not find begin label
      if((mFilePos>=startpos) && (!firstgo)) {
	std::stringstream errstr;
	errstr << "Section \"" << rLabel << "\" expected at " << FileLine() << ", did not find begin lable";
	throw Exception("TokenReader::ReadBegin Missed", errstr.str(), 51);
      }
      // skip this token
      Get(token);
    }
  }
  // catch my seek/tell errors
  catch (std::ios::failure&) {
    std::stringstream errstr;
    errstr << "Section \"" << rLabel << "\" expected at " << FileLine();
    throw Exception("TokenReader::ReadBegin Rewind", errstr.str(), 1);
  }
}
 

// ReadEnd(label)
void TokenReader::ReadEnd(const std::string& rLabel) {
  FD_DV("TokenReader::ReadEnd: " << rLabel << " at " << FileLine() );
  // search for end at current level
  int level=mLevel;
  Token token;
  for (;;) {
    // exception: did not get a token at all
    if(!Get(token)) {
      std::stringstream errstr;
      errstr << "end of Section \"" << rLabel << "\" expected at " << FileLine();
      throw Exception("TokenReader::ReadEnd", errstr.str(), 51);
    }
    // success: found end of current section
    if ((token.Type() == Token::End) && (token.StringValue() == rLabel) && (mLevel==level-1)) {
      break;
    }
    // exception: current section ends with unexpected label
    if(mLevel<level) {
      std::stringstream errstr;
      errstr << "end of Section \"" << rLabel << "\" expected at " << FileLine();
      throw Exception("TokenReader::ReadEnd", errstr.str(), 51);
    }
  }
}

// Eos(label)
bool TokenReader::Eos(const std::string& rLabel) {
  // peek token and check for end of section
  Token token;
  Peek(token);
  if (token.Type() != Token::End) 
    return false;
  if ((token.Type() == Token::End) && (token.StringValue() == rLabel))
    return true; 
  std::stringstream errstr;
  errstr << "Section End\"" << rLabel << "\" expected at " << FileLine();
  throw Exception("TokenReader::Eos", errstr.str(), 51);
  return false;
}
  

// ReadInteger()
long int TokenReader::ReadInteger(void) {
  Token token;			
  Get(token);
  if((token.Type()!=Token::Integer) && (token.Type()!=Token::Integer16)  ) {
    std::stringstream errstr;
    errstr << "Integer expected at " << FileLine();
    throw Exception("TokenReader::TokenReader", errstr.str(), 50);
  }
  return token.IntegerValue();
}

// ReadFloat()
double TokenReader::ReadFloat(void) {
  Token token;			
  Get(token);
  if((token.Type()!=Token::Float) && (token.Type()!=Token::Integer)) {
    std::stringstream errstr;
    errstr << "Float expected at " << FileLine();
    throw Exception("TokenReader::TokenReader", errstr.str(), 50);
  }
  return token.FloatValue();
}

// ReadString()
const std::string& TokenReader::ReadString(void) {
  Token token;			
  Get(token);
  if(token.Type()!=Token::String) {
    std::stringstream errstr;
    errstr << "Name expected at " << FileLine();
    throw Exception("TokenReader::TokenReader", errstr.str(), 50);
  }
  mLastString=token.StringValue();
  return(mLastString);
}

 
// ReadOption()
const std::string& TokenReader::ReadOption(void) {
  Token token;			
  Get(token);
  if(token.Type()!=Token::Option) {
    std::stringstream errstr;
    errstr << "Option expected at " << FileLine();
    throw Exception("TokenReader::TokenReader", errstr.str(), 50);
  }
  mLastString=token.StringValue();
  return(mLastString);
}

// ReadBinary()
const std::string& TokenReader::ReadBinary(void) {
  Token token;			
  Get(token);
  if(token.Type()!=Token::Binary) {
    std::stringstream errstr;
    errstr << "Binary string expected at " << FileLine();
    throw Exception("TokenReader::TokenReader", errstr.str(), 50);
  }
  mLastString=token.StringValue();
  return(mLastString);
}


 

// Line()
int TokenReader::Line(void) const {
  return mLineCount;
}

// FileLine()
std::string TokenReader::FileLine(void) const {
  if(mFileName!="")
    return "("+ mFileName + ":" + ToStringInteger(mLineCount) +")";
  else
    return "(#" + ToStringInteger(mLineCount) +")";
}

} // namespace faudes
