/** @file tokenwriter.cpp @brief Class TokenWriter */

/* FAU Discrete Event Systems Library (libfaudes)

Copyright (C) 2006  Bernd Opitz
Copyright (C) 2006  Thomas Moor
Exclusive copyright is granted to Klaus Schmidt

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#include "tokenwriter.h"

namespace faudes {

// TokenWriter(rFilename, openmode)
TokenWriter::TokenWriter(const std::string& rFilename, std::ios::openmode openmode) 
  : mMode(File), mFileName(rFilename), mColumns(80/FD_NAMELEN), mColCount(0) {
  // set up mFStream
  mFStream.exceptions(std::ios::badbit|std::ios::failbit);
  try{
    mFStream.open(rFilename.c_str(), openmode); 
  }
  catch (std::ios::failure&) {
    std::stringstream errstr;
    errstr << "Exception opening/writing file \""<< rFilename << "\"";
    throw Exception("TokenWriter::TokenWriter", errstr.str(), 2);
  }
  // use mFStream
  mpStream=&mFStream;      
  mEndl=true;
}

// TokenWriter(mode)
TokenWriter::TokenWriter(Mode mode)
  : mMode(mode), mFileName(""), mColumns(80/FD_NAMELEN), mColCount(0) {
  switch(mode) {
  case Stdout: 
    // set up mFStream
    mFileName="stdout";
    /*
      try {
      mFStream.copyfmt(std::cout);
      mFStream.clear(std::cout.rdstate());
      typedef std::basic_ios<char> ___basic_ios_char_; // trick for vc++
      mFStream.___basic_ios_char_::rdbuf(std::cout.rdbuf());
      }
      catch (std::ios::failure&) {
      std::stringstream errstr;
      errstr << "Exception opening/writing file \""<< mFileName << "\"";
      throw Exception("TokenWriter::TokenWriter", errstr.str(), 2);
      }
      // use mFStream
      mpStream=&mFStream;
    */
    // use std::cout
    mpStream= &std::cout;
    mEndl=true;
    break;
  case String:
    // use mSStream
    mFileName="string";
    mpStream=&mSStream;
    mEndl=false;
    break;
  default:
    std::stringstream errstr;
    errstr << "Invalid Mode / Not Implemented";
    throw Exception("TokenWriter::TokenWriter", errstr.str(), 2);
  }
}


// destructor
TokenWriter::~TokenWriter(void) {
  if(mMode==File) mFStream.close();
}

// Str()
std::string TokenWriter::Str(void) {
  if(mMode!=String) {
    std::stringstream errstr;
    errstr << "Not in String Mode";
    throw Exception("TokenWriter::Str()", errstr.str(), 2);
  }
  return mSStream.str();
}

// Columns()
int TokenWriter::Columns(void) const {
  return mColumns;
}

// Columns()
void TokenWriter::Columns(int columns) {
  mColumns = columns;
}

// Endl()
void TokenWriter::Endl(void) {
  try{
    if(mEndl) *mpStream << std::endl;
    else *mpStream << " ";
  }
  catch (std::ios::failure&) {
    std::stringstream errstr;
    errstr << "Exception opening/writing file \"" << mFileName << "\"";
    throw Exception("Generator::write", errstr.str(), 2);
  }
}

// Endl(bool)
void TokenWriter::Endl(bool on) {
  mEndl=on;
}


// Write(rToken)
void TokenWriter::Write(Token& rToken) {
  FD_DV("TokenWriter::Write(token)");
  // if(mMode==Stdout) mpStream = &std::cout; // fix
  try{
    if ( (rToken.Type() == Token::Begin) || (rToken.Type() == Token::End) ) {
      if(mColCount !=0) Endl();
      rToken.Write(mpStream);
      Endl();
      mColCount = 0;
    }
    else {
      if(mColCount == mColumns) {
	mColCount = 0;
	Endl();
      }
      mColCount++;
      rToken.Write(mpStream);
      *mpStream << " ";
    }
  }
  catch (std::ios::failure&) {
    std::stringstream errstr;
    errstr << "Exception opening/writing file \"" << mFileName << "\"";
    throw Exception("TokenWriter::Write(token)", errstr.str(), 2);
  }
}

// WriteString(rName)
void TokenWriter::WriteString(const std::string& rName) {
  if ((rName == "\n") || (rName == "\r\n")) {
    Endl();
    mColCount = 0;
    return;
  }
  Token token;
  token.SetString(rName);
  Write(token);
}


// WriteInteger(index)
void TokenWriter::WriteInteger(Idx index) {
  Token token;
  token.SetInteger(index);
  Write(token);
}

// WriteInteger(index)
void TokenWriter::WriteInteger16(long int val) {
  Token token;
  token.SetInteger16(val);
  Write(token);
}

// WriteFloat(float)
void TokenWriter::WriteFloat(const double& val) {
  Token token;
  token.SetFloat(val);
  Write(token);
}


// WriteOption(rOpt)
void TokenWriter::WriteOption(const std::string& rOpt) {
  Token token;
  token.SetOption(rOpt);
  Write(token);
}


// WriteBegin(rLabel)
void TokenWriter::WriteBegin(const std::string& rLabel) {
  Token token;
  token.SetBegin(rLabel);
  Write(token);
}

// WriteEnd(rLabel)
void TokenWriter::WriteEnd(const std::string& rLabel) {
  Token token;
  token.SetEnd(rLabel);
  Write(token);
}

// WriteComment(comment)
void TokenWriter::WriteComment(const std::string& comment){
  if(!mEndl) return; //  no endl implies no comments
  try{
    if(mColCount!=0)  Endl();
    if(comment.length()==0) 
      *mpStream << "% ";
    if(comment.length()>0) 
    if(comment.at(0)!='%')
      *mpStream << "% ";
    *mpStream << comment;
    Endl();
    mColCount = 0;
  }
  catch (std::ios::failure&) {
    std::stringstream errstr;
    errstr << "Exception opening/writing file \"" << mFileName << "\"";
    throw Exception("TokenWriter::Comment", errstr.str(), 2);
  }
}

// write base64
void TokenWriter::WriteBinary(const char* pData, long int len) {
  Token::WriteBinary(mpStream,pData,len);
}

// FileName()
std::string TokenWriter::FileName(void) const {
  return mFileName;
}




} // namespace faudes
